<?php

namespace OtomaticAi\Api\Flux;

use Exception;
use OtomaticAi\Api\Ideogram\Exceptions\ApiException;
use OtomaticAi\Models\Usages\Usage;
use OtomaticAi\Utils\Settings;
use OtomaticAi\Vendors\GuzzleHttp\Client as HttpClient;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\GuzzleHttp\Pool;
use OtomaticAi\Vendors\GuzzleHttp\Psr7\Request;
use OtomaticAi\Vendors\GuzzleHttp\Psr7\Response;

class PoolClient
{
    /**
     * The Flux api endpoint
     *
     * @var string
     */
    private string $endpoint = "https://api.bfl.ml/v1";

    /**
     * The GuzzleHttp client
     *
     * @var HttpClient
     */
    private HttpClient $client;

    /**
     * The auth key
     *
     * @var HttpClient
     */
    private string $key;

    // cache for store tasks data
    public array $tasks = [];

    /**
     * Create a new Flux Api client
     *
     * @param string|null $key
     * @throws Exception
     */
    function __construct(string $key = null)
    {
        // get the key
        if (empty($key)) {
            $key = Settings::get("api.flux.api_key");
        }
        if (empty($key)) {
            throw new Exception("No Flux Api Key provided.");
        }

        $this->key = $key;

        // create the http client
        $this->client = new HttpClient(
            [
                "base_uri" => rtrim($this->endpoint, "/") . '/',
                "headers" => [
                    "X-Key" => $this->key,
                    "Accept" => "application/json",
                    "Content-Type" => "application/json",
                ],
                "timeout" => 600
            ]
        );
    }

    /**
     * Call the create task api endpoint
     *
     * @param array $settings
     * @return array
     * @throws Exception
     */
    public function createTasks(array $settings = []): array
    {
        $requests = [];

        foreach ($settings as $payload) {
            // make the payload
            $payload = array_merge([
                "width" => 1344,
                "height" => 768,
                "prompt_upsampling" => true,
            ], $payload);

            // make the request
            $requests[] = [
                "request" => new Request(
                    'POST',
                    $payload["model"],
                    [],
                    json_encode($payload)
                ),
                "after" => function ($result) use ($payload) {

                    $id = Arr::get($result, "id");

                    $this->tasks[$id] = [
                        "id" => $id,
                        "payload" => $payload,
                        "result" => null,
                    ];
                }
            ];
        }

        // call the api endpoint
        $responses = $this->run($requests);

        return $responses;
    }

    /**
     * Call the get result api endpoint
     *
     * @param array $settings
     * @return array
     * @throws Exception
     */
    public function getResults(array $ids): array
    {
        $requests = [];
        foreach ($ids as $id) {
            $status = Arr::get($this->tasks, $id . ".result.status", null);
            if (empty($status) || $status === "Pending") {
                $requests[] = [
                    "request" => new Request(
                        'GET',
                        "get_result?id=" . $id,
                    )
                ];
            }
        }

        // call the api endpoint
        $responses = $this->run($requests);

        foreach ($responses as $response) {
            if (isset($this->tasks[$response["id"]])) {
                $this->tasks[$response["id"]]["result"] = $response;
            }
        }

        $output = [];
        foreach ($ids as $id) {
            $output[] = Arr::get($this->tasks, $id . ".result");
        }

        return $output;
    }

    /**
     * Call the generate api endpoint
     *
     * @param array $settings
     * @return array
     * @throws Exception
     */
    public function generate(array $settings = []): array
    {
        $requests = [];
        foreach ($settings as $payload) {

            // make the payload
            $payload = [
                "image_request" => array_merge([
                    "model" => "V_2",
                    "magic_prompt_option" => "AUTO",
                    "aspect_ratio" => "ASPECT_16_9",
                ], Arr::get($payload, "image_request", []))
            ];

            // make the request
            $requests[] = [
                "request" => new Request(
                    'POST',
                    "generate",
                    [],
                    json_encode($payload)
                ),
                "after" => function ($complete) use ($payload) {

                    // store the usage
                    Usage::create([
                        "provider" => "ideogram",
                        "payload" => [
                            "engine" => $payload["image_request"]["model"],
                            "artifacts" => 1,
                        ]
                    ]);
                }
            ];
        }

        // call the stability ai api
        $completes = $this->run($requests);

        return $completes;
    }

    /**
     * Perform an api request
     *
     * @param array $requests
     * @param stirng $uri
     * @param array $options
     * @return array
     * @throws Exception
     */
    private function run(array $requests)
    {
        $output = [];

        // get the http requests
        $httpRequests = Arr::pluck($requests, "request");

        // create the pool
        $pool = new Pool($this->client, $httpRequests, [
            'concurrency' => 10,
            'fulfilled' => function (Response $response, $index) use (&$output, $requests) {
                $response = $response->getBody()->getContents();
                $response = json_decode($response, true);
                $output[$index] = $response;

                // call after callback
                if (!empty($requests[$index]["after"]) && is_callable($requests[$index]["after"])) {
                    $requests[$index]["after"]($response);
                }
            },
            'rejected' => function (Exception $reason, $index) {
                $e = ApiException::make($reason);
                throw $e;
            },
        ]);

        // call the pool
        $promise = $pool->promise();
        $promise->wait();

        ksort($output);
        $output = array_values($output);

        return $output;
    }
}
