<?php

namespace OtomaticAi\Controllers;

use Exception;
use OtomaticAi\Api\OpenAi\Client as OpenAiClient;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Api\StabilityAi\Client as StabilityAiClient;
use OtomaticAi\Api\Ideogram\Client as IdeogramClient;
use OtomaticAi\Api\Flux\Client as FluxClient;
use OtomaticAi\Models\Preset;
use OtomaticAi\Utils\Auth;
use OtomaticAi\Utils\Image;
use OtomaticAi\Utils\Language;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;
use OtomaticAi\Vendors\Illuminate\Validation\Rule;

class ImageGeneratorController extends Controller
{
    public function generate()
    {
        $this->verifyNonce();

        $this->validate([
            "provider" => ["required", Rule::in('dall_e', 'stable_diffusion', 'ideogram', 'flux')],
            "prompt" => ["required", "string"],
        ]);

        // refresh auth
        Auth::refreshDomain();

        if (!Auth::isPremium() && !Auth::isTrial()) {
            $this->response(["message" => "An error occurred", "error" => "This feature requires a premium subscription."], 503);
        }

        // save start event
        $this->collectEvent(
            "generate-image-start",
            [
                "provider" => $this->input("provider")
            ]
        );

        $prompt = $this->translatePrompt($this->input("prompt"));

        try {
            $image = null;
            switch ($this->input("provider")) {
                case "dall_e":
                    $image = $this->generateDallEImage($prompt, $this->input("settings.dall_e"));
                    break;
                case "stable_diffusion":
                    $image = $this->generateStableDiffusionImage($prompt, $this->input("settings.stable_diffusion"));
                    break;
                case "ideogram":
                    $image = $this->generateIdeogramImage($prompt, $this->input("settings.ideogram"));
                    break;
                case "flux":
                    $image = $this->generateFluxImage($prompt, $this->input("settings.flux"));
                    break;
            }

            if (!empty($image)) {

                // refresh auth
                Auth::refreshDomain();

                // save end event
                $this->collectEvent("generate-image-end",  [
                    "provider" => $this->input("provider")
                ]);

                $this->response($image);
            }
        } catch (Exception $e) {

            $this->collectEvent("generate-image-failed",  [
                "provider" => $this->input("provider"),
                "message" => $e->getMessage(),
            ]);

            $this->response(["message" => "An error occurred", "error" => $e->getMessage()], 503);
        }

        $this->response(["message" => "An error occurred", "error" => "No image generated."], 503);
    }

    public function save()
    {
        $this->verifyNonce();

        $this->validate([
            "name" => ["nullable"],
            "description" => ["nullable"],
            "legend" => ["nullable"],
            "post_id" => ["nullable"],
            "path" => ["required", "string"],
        ]);

        $image = Image::fromPath($this->input("path"));
        if (($attachementId = $image->save($this->input("name"), $this->input("description"), $this->input("legend"))) !== null) {
            if (!empty($this->input("post_id"))) {
                if (set_post_thumbnail($this->input("post_id"), $attachementId) === false) {
                }
            }
        }

        $this->response([]);
    }

    /**
     * Generate an image description from title
     *
     */
    public function generateDescription()
    {
        $this->verifyNonce();

        $this->validate([
            "title" => ["required", "string"],
        ]);

        try {
            $preset = Preset::findFromAPI("generate_alt_image");

            $response = $preset->process([
                "language" => Language::findFromLocale()->value,
                "request" => $this->input("title"),
            ]);

            // get the response content
            $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
            $content = Arr::get($response, "value");
            $content = Str::clean($content);

            $this->response([
                "description" => Str::lower($content),
            ]);
        } catch (Exception $e) {
        }

        return $this->input("title");
    }

    /**
     * Translate the prompt in english
     *
     * @param string $prompt
     * @return string
     */
    private function translatePrompt(string $prompt): string
    {
        try {
            // get the openai preset
            $preset = Preset::findFromAPI("translate_content");

            // make payload
            $payload = [
                "language" => Language::find("en")->value,
                "content" => $prompt,
            ];


            // run the preset
            $response = $preset->process($payload);

            // get the response content
            $content = Arr::get($response, 'choices.0.message.content');
            $content = Str::clean($content);

            // return translated content
            if (!empty($content)) {
                return $content;
            }
        } catch (Exception $e) {
        }

        // return prompt if translation was incorrect
        return $prompt;
    }

    /**
     * save the base64 image to temp folder
     *
     * @param string $base64
     * @return string|null
     */
    private function imageUrlToBase64(string $url): ?string
    {
        $mediaFile = wp_remote_get($url);

        if (is_wp_error($mediaFile)) {
            throw new Exception("Can not download image url `" . $url . "`. Error: " . $mediaFile->get_error_message());
        }

        return base64_encode($mediaFile["body"]);
    }

    /**
     * save the base64 image to temp folder
     *
     * @param string $base64
     * @return string|null
     */
    private function saveImageInTemp(string $base64): ?string
    {
        $extension = "png";
        $name = uniqid();
        $path = get_temp_dir() . $name . '.' . $extension;

        $base64 = base64_decode($base64);

        if (file_put_contents($path, $base64) !== false) {
            return $path;
        }
    }

    private function collectEvent(string $key, array $payload = [])
    {
        $api = new Client;
        $api->collectEvent($key, $payload);
    }

    /**
     * Generate a stable diffusion image
     *
     * @param string $prompt
     * @param array $settings
     * @return array|null
     */
    private function generateStableDiffusionImage(string $prompt, array $settings = []): ?array
    {
        $api = new StabilityAiClient();

        switch ($settings["model"]) {
            case "sdxl":
                // sdxl model
                $payload = [
                    "text_prompts" => [
                        [
                            "text" => $prompt,
                            "weight" => 1
                        ]
                    ]
                ];

                $result = $api->textToImage($payload);

                if (!empty(Arr::get($result, "artifacts.0.base64"))) {
                    $path = $this->saveImageInTemp(Arr::get($result, "artifacts.0.base64"));

                    return [
                        "path" => $path,
                        "provider" => "stable_diffusion",
                        "image" => [
                            "base64" => Arr::get($result, "artifacts.0.base64"),
                            "height" => 768,
                            "width" => 1344,
                        ]
                    ];
                }
                break;
            default:
                // others models
                $payload = [
                    "prompt" => $prompt,
                ];

                switch ($settings["model"]) {
                    case "stable_image_core":
                        $payload["model"] = "core";
                        break;
                    case "stable_image_sd3_medium":
                        $payload["model"] = "sd3-medium";
                        break;
                    case "stable_image_sd3_large_turbo":
                        $payload["model"] = "sd3-large-turbo";
                        break;
                    case "stable_image_sd3_large":
                        $payload["model"] = "sd3-large";
                        break;
                    case "stable_image_sd3_5_large":
                        $payload["model"] = "sd3.5-large";
                        break;
                    case "stable_image_sd3_5_large_turbo":
                        $payload["model"] = "sd3.5-large-turbo";
                        break;
                    case "stable_image_ultra":
                        $payload["model"] = "ultra";
                        break;
                    default:
                        $payload["model"] = "core";
                }

                $result = $api->stableImageGenerate($payload);

                if (!empty(Arr::get($result, "image"))) {
                    $path = $this->saveImageInTemp(Arr::get($result, "image"));

                    if (!empty($path)) {
                        return [
                            "path" => $path,
                            "provider" => "stable_diffusion",
                            "image" => [
                                "base64" => Arr::get($result, "image"),
                                "height" => 768,
                                "width" => 1344,
                            ]
                        ];
                    }
                }
        }

        return null;
    }

    /**
     * Generate a ideogram image
     *
     * @param string $prompt
     * @param array $settings
     * @return array|null
     */
    private function generateIdeogramImage(string $prompt, array $settings = [])
    {
        $api = new IdeogramClient();

        $payload = [
            "image_request" => [
                "prompt" => $prompt,
            ]
        ];

        switch ($settings["model"]) {
            case "ideogram_v_1":
                $payload["image_request"]["model"] = "V_1";
                break;
            case "ideogram_v_1_turbo":
                $payload["image_request"]["model"] = "V_1_TURBO";
                break;
            case "ideogram_v_2":
                $payload["image_request"]["model"] = "V_2";
                break;
            case "ideogram_v_2_turbo":
                $payload["image_request"]["model"] = "V_2_TURBO";
                break;
            default:
                $payload["image_request"]["model"] = "V_2";
        }

        $result = $api->generate($payload);

        if (!empty(Arr::get($result, "data.0.url"))) {
            $base64 = $this->imageUrlToBase64(Arr::get($result, "data.0.url"));
            $path = $this->saveImageInTemp($base64);

            if (!empty($path)) {
                return [
                    "path" => $path,
                    "provider" => "ideogram",
                    "image" => [
                        "base64" => $base64,
                        "width" => $settings["model"] === "dall-e-3" ? 1792 : 1024,
                        "height" => 1024,
                    ]
                ];
            }
        }

        return null;
    }

    /**
     * Generate a Flux image
     *
     * @param string $prompt
     * @param array $settings
     * @return array|null
     */
    private function generateFluxImage(string $prompt, array $settings = [])
    {
        $api = new FluxClient();

        $payload = [
            "prompt" => $prompt,
        ];

        switch ($settings["model"]) {
            case "flux_1_dev":
                $payload["model"] = "flux-dev";
                break;
            case "flux_1_pro":
                $payload["model"] = "flux-pro";
                break;
            case "flux_1_1_pro":
                $payload["model"] = "flux-pro-1.1";
                break;
            default:
                $payload["model"] = "flux-dev";
        }

        $result = $api->createTask($payload);

        $id = Arr::get($result, "id");
        if (!empty($id)) {

            while (true) {

                // wait 3 seconds for the task to be ready
                sleep(3);

                $result = $api->getResult($id);

                switch (Arr::get($result, "status")) {
                    case "Pending":
                        break;
                    case "Ready":
                        $base64 = $this->imageUrlToBase64(Arr::get($result, "result.sample"));
                        $path = $this->saveImageInTemp($base64);

                        if (!empty($path)) {
                            return [
                                "path" => $path,
                                "provider" => "flux",
                                "image" => [
                                    "base64" => $base64,
                                    "width" => 1344,
                                    "height" => 1024,
                                ]
                            ];
                        }
                        break;
                    default:
                        throw new Exception("Flux image generation failed. Status: " . Arr::get($result, "status"));
                }
            }
        }

        return null;
    }

    /**
     * Generate a dall-e image
     *
     * @param string $prompt
     * @param array $settings
     * @return array|null
     */
    private function generateDallEImage(string $prompt, array $settings = [])
    {
        $api = new OpenAiClient();

        $payload = [
            "model" => $settings["model"],
            "quality" => "standard",
            "size" => $settings["model"] === "dall-e-3" ? "1792x1024" : "1024x1024",
        ];

        $result = $api->image($prompt, $payload);

        if (!empty(Arr::get($result, "data.0.b64_json"))) {
            $path = $this->saveImageInTemp(Arr::get($result, "data.0.b64_json"));

            if (!empty($path)) {
                return [
                    "path" => $path,
                    "provider" => "dall_e",
                    "image" => [
                        "base64" => Arr::get($result, "data.0.b64_json"),
                        "width" => $settings["model"] === "dall-e-3" ? 1792 : 1024,
                        "height" => 1024,
                    ]
                ];
            }
        }

        return null;
    }
}
