<?php

namespace OtomaticAi\Controllers\SingleMode;

use DOMDocument;
use DOMXPath;
use OtomaticAi\Api\OpenAi\Client as OpenAiClient;
use OtomaticAi\Api\StabilityAi\Client as StabilityAiClient;
use OtomaticAi\Api\Ideogram\Client as IdeogramClient;
use OtomaticAi\Api\Flux\Client as FluxClient;
use OtomaticAi\Api\Unsplash\Client as UnsplashClient;
use OtomaticAi\Api\Pexels\Client as PexelsClient;
use OtomaticAi\Api\Pixabay\Client as PixabayClient;
use OtomaticAi\Api\GoogleImage\Client as GoogleImageClient;
use OtomaticAi\Api\BingImage\Client as BingImageClient;
use OtomaticAi\Content\Image\BingImage;
use OtomaticAi\Content\Image\DallE;
use OtomaticAi\Content\Image\Flux;
use OtomaticAi\Content\Image\GoogleImage;
use OtomaticAi\Content\Image\Ideogram;
use OtomaticAi\Content\Image\Image;
use OtomaticAi\Content\Image\Pexels;
use OtomaticAi\Content\Image\Pixabay;
use OtomaticAi\Content\Image\StableDiffusion;
use OtomaticAi\Content\Image\Unsplash;
use OtomaticAi\Controllers\Controller;
use OtomaticAi\Models\SingleModeGeneration;
use OtomaticAi\Vendors\Illuminate\Support\Arr;

class MediaController extends Controller
{
    public function regenerateImage()
    {
        $this->validate([
            "id" => ["required", "integer"],
            "media_id" => ["required", "string"],
        ]);

        $model = SingleModeGeneration::find($this->input('id'));

        if (empty($model)) {
            $this->response(["message" => "An error occurred", "error" => "Unable to find the generation #" . $this->input("id") . "."], 503);
        }

        // load html
        $doc = new DOMDocument('1.0', 'UTF-8');
        libxml_use_internal_errors(true);
        $doc->loadHTML('<?xml encoding="UTF-8">' . $model->content, LIBXML_NOWARNING);
        libxml_use_internal_errors(false);

        // create xpath
        $xpath = new DOMXPath($doc);

        $nodes = $xpath->query("//otoimage[@uid='" . $this->input('media_id') . "']");

        if ($nodes->length > 0) {
            $node = $nodes->item(0);
            $image = Image::fromHtmlElement($node);

            if (!empty($image)) {
                $newImage = null;
                switch ($image::PROVIDER) {
                    case "stable-diffusion":
                        $newImage = $this->getStableDiffusionImage($image, $model);
                        break;
                    case "ideogram":
                        $newImage = $this->getIdeogramImage($image, $model);
                        break;
                    case "flux":
                        $newImage = $this->getFluxImage($image, $model);
                        break;
                    case "dall-e":
                        $newImage = $this->getDallEImage($image, $model);
                        break;
                    case "unsplash":
                        $newImage = $this->getUnsplashImage($image, $model);
                        break;
                    case "pexels":
                        $newImage = $this->getPexelsImage($image, $model);
                        break;
                    case "pixabay":
                        $newImage = $this->getPixabayImage($image, $model);
                        break;
                    case "google-image":
                        $newImage = $this->getGoogleImage($image, $model);
                        break;
                    case "bing-image":
                        $newImage = $this->getBingImage($image, $model);
                        break;
                }

                if (!empty($newImage)) {
                    $node->parentNode->replaceChild($newImage->toHtmlElement($doc), $node);
                }
            }
        }

        $model->content = trim(preg_replace("/^<body>|<\/body>$/", "", $doc->saveHTML($doc->getElementsByTagName("body")->item(0))));
        $model->save();

        $this->response($model->fresh());
    }

    /**
     * Generate a stable diffusion image
     *
     * @param StableDiffusion $image
     * @param SingleModeGeneration $generation
     * @return StableDiffusion|StableDiffusion[]|null
     */
    private function getStableDiffusionImage(StableDiffusion $image, SingleModeGeneration $generation)
    {
        // make the payload
        if ($image->model === "sdxl") {
            $payload = [
                "text_prompts" => [
                    [
                        "text" => $image->prompt,
                        "weight" => 1
                    ]
                ]
            ];
        } else {
            // others models
            $payload = [
                "prompt" => $image->prompt,
            ];

            switch ($image->model) {
                case "stable_image_core":
                    $payload["model"] = "core";
                    break;
                case "stable_image_sd3_medium":
                    $payload["model"] = "sd3-medium";
                    break;
                case "stable_image_sd3_turbo":
                case "stable_image_sd3_large_turbo":
                    $payload["model"] = "sd3-large-turbo";
                    break;
                case "stable_image_sd3":
                case "stable_image_sd3_large":
                    $payload["model"] = "sd3-large";
                    break;
                case "stable_image_sd3_5_large":
                    $payload["model"] = "sd3.5-large";
                    break;
                case "stable_image_sd3_5_large_turbo":
                    $payload["model"] = "sd3.5-large-turbo";
                    break;
                case "stable_image_ultra":
                    $payload["model"] = "ultra";
                    break;
                default:
                    $payload["model"] = "core";
            }
        }

        // call the stability ai api
        $api = new StabilityAiClient();
        if ($image->model === "sdxl") {
            // sdxl model
            $result = [];
            $result = $api->textToImage($payload);
            if (Arr::get($result, "artifacts.0.finishReason") === "SUCCESS") {
                return StableDiffusion::make(Arr::get($result, "artifacts.0.base64"), $generation->title, $image->description, null,  $image->prompt, "sdxl");
            }
        } else {
            // others models
            $result = $api->stableImageGenerate($payload);

            if (!empty(Arr::get($result, "image"))) {
                return StableDiffusion::make(Arr::get($result, "image"), $generation->title, $image->description, null,  $image->prompt, $image->model);
            }
        }
    }

    /**
     * Generate a ideogram image
     *
     * @param Ideogram $image
     * @param SingleModeGeneration $generation
     * @return Ideogram|Ideogram[]|null
     */
    private function getIdeogramImage(Flux $image, SingleModeGeneration $generation)
    {
        // make the payload
        $payload = [
            "image_request" => [
                "prompt" => $image->prompt,
            ]
        ];

        switch ($image->model) {
            case "ideogram_v_1":
                $payload["image_request"]["model"] = "V_1";
                break;
            case "ideogram_v_1_turbo":
                $payload["image_request"]["model"] = "V_1_TURBO";
                break;
            case "ideogram_v_2":
                $payload["image_request"]["model"] = "V_2";
                break;
            case "ideogram_v_2_turbo":
                $payload["image_request"]["model"] = "V_2_TURBO";
                break;
            default:
                $payload["image_request"]["model"] = "V_2";
        }


        // call the ideogram api
        $api = new IdeogramClient();
        $result = $api->generate($payload);

        // make the Ideogram Image
        if (!empty(Arr::get($result, "data.0.url", null))) {
            return Ideogram::make(Arr::get($result, "data.0.url"), $generation->title, $image->description, null, Arr::get($result, "data.0.prompt"));
        }
    }

    /**
     * Generate a Flux image
     *
     * @param Flux $image
     * @param SingleModeGeneration $generation
     * @return Flux|Flux[]|null
     */
    private function getFluxImage(Flux $image, SingleModeGeneration $generation)
    {
        // make the payload
        $payload = [
            "prompt" => $image->prompt,
            "model" => "flux-dev",
        ];

        switch ($image->model) {
            case "flux_1_dev":
                $payload["model"] = "flux-dev";
                break;
            case "flux_1_pro":
                $payload["model"] = "flux-pro";
                break;
            case "flux_1_1_pro":
                $payload["model"] = "flux-pro-1.1";
                break;
            default:
                $payload["model"] = "flux-dev";
        }

        $api = new FluxClient();

        // create the tasks
        $response = $api->createTask($payload);

        // get the task id
        $id = Arr::get($response, "id");

        // get the tasks results
        $result = null;
        while (true) {
            sleep(3);

            $result = $api->getResult($id);
            if (Arr::get($result, "status") !== "Pending") {
                break;
            }
        }

        // make the Flux Image
        if (!empty($result) && !empty(Arr::get($result, "result.sample"))) {
            return Flux::make(Arr::get($result, "result.sample"), $generation->title, $image->description, null, Arr::get($result, "result.prompt"), $payload["model"]);
        }
    }

    /**
     * Generate a dall-e image
     *
     * @param DallE $image
     * @param SingleModeGeneration $generation
     * @return DallE|DallE[]|null
     */
    private function getDallEImage(DallE $image, SingleModeGeneration $generation)
    {


        $prompts = [];



        // make the pool settings
        $settings = [
            "prompt" => [],
            "settings" => [],
        ];

        $settings["prompt"] = $image->prompt;
        $model = $image->model;
        $settings["settings"] = [
            "model" => $model,
            "quality" => "standard",
            "size" => $model === "dall-e-3" ? "1792x1024" : "1024x1024",
        ];

        // call the openai api
        $api = new OpenAiClient();
        $result = $api->image($settings["prompt"], $settings["settings"]);

        // make the DallE Image
        if (!empty($result) && !empty(Arr::get($result, "data.0.b64_json"))) {
            return DallE::make(Arr::get($result, "data.0.b64_json"), $generation->title, $image->description, null, Arr::get($result, "data.0.revised_prompt"));
        }
    }

    /**
     * Generate an Unsplash image
     *
     * @param Unsplash $image
     * @param SingleModeGeneration $generation
     * @return Unsplash|Unsplash[]|null
     */
    private function getUnsplashImage(Unsplash $image, SingleModeGeneration $generation)
    {

        // make the search for unsplash
        $search = $image->search;

        // call unsplash api
        $api = new UnsplashClient;
        $response = $api->searchPhotos([
            "query" => $search,
            "per_page" => 20,
        ]);

        // get results
        $result = Arr::get($response, "results", []);

        if (!empty($result)) {

            // get a random image
            $result = Arr::random($result);

            // make the Unsplash Image
            return Unsplash::make($result["urls"]["regular"], $generation->title, $image->description, null, $search);
        }
    }

    /**
     * Generate an Pexels image
     *
     * @param Pexels $image
     * @param SingleModeGeneration $generation
     * @return Pexels|Pexels[]|null
     */
    private function getPexelsImage(Pexels $image, SingleModeGeneration $generation)
    {
        // make the search for pexels
        $search = $image->search;

        // call pexels api
        $api = new PexelsClient();
        $response = $api->searchPhotos([
            "query" => $search,
            "per_page" => 20,
        ]);

        // get results
        $result = Arr::get($response, "photos", []);

        if (!empty($result)) {

            // get a random image
            $result = Arr::random($result);

            // make the Pexels Image
            return Pexels::make($result["src"]["landscape"], $generation->title, $image->description, null, $search);
        }
    }

    /**
     * Generate an Pixabay image
     *
     * @param Pixabay $image
     * @param SingleModeGeneration $generation
     * @return Pixabay|Pixabay[]|null
     */
    private function getPixabayImage(Pixabay $image, SingleModeGeneration $generation)
    {
        // make the search for pixabay
        $search = $image->search;

        // call pixabay api
        $api = new PixabayClient();
        $response = $api->images([
            "q" => $search,
        ]);

        // get a random image
        $result = Arr::get($response, "hits", []);

        if (!empty($result)) {

            // get a random image
            $result = Arr::random($result);

            // make the Pixabay Image
            return Pixabay::make($result["largeImageURL"], $generation->title, $image->description, null, $search);
        }
    }

    /**
     * Generate an Google image
     *
     * @param GoogleImage $image
     * @param SingleModeGeneration $generation
     * @return GoogleImage|GoogleImage[]|null
     */
    private function getGoogleImage(GoogleImage $image, SingleModeGeneration $generation)
    {
        // make the search for google image
        $search = $image->search;

        // call google image api
        $api = new GoogleImageClient();
        $result = $api->images([
            "q" => $search,
        ]);

        if (!empty($result)) {

            // get a random image
            $result = Arr::random($result);

            // make the Google Image
            return GoogleImage::make($result, $generation->title, $image->description, null, $search);
        }
    }

    /**
     * Generate an Bing image
     *
     * @param BingImage $image
     * @param SingleModeGeneration $generation
     * @return BingImage|BingImage[]|null
     */
    private function getBingImage(BingImage $image, SingleModeGeneration $generation)
    {
        // make the search for bing image
        $search = $image->search;

        // call google image api
        $api = new BingImageClient();
        $result = $api->images([
            "q" => $search,
        ]);

        if (!empty($result)) {

            // get a random image
            $result = Arr::random($result);

            // make the Bing Image
            return BingImage::make($result, $generation->title, $image->description, null, $search);
        }
    }
}
