<?php

namespace OtomaticAi\Controllers;

use Exception;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Models\Project;
use OtomaticAi\Models\Template;
use OtomaticAi\Vendors\Carbon\Carbon;
use OtomaticAi\Vendors\Illuminate\Support\Arr;

class TemplateController extends Controller
{
    public function index()
    {
        $this->verifyNonce();

        $this->response([
            "local" => $this->getLocalTemplates(),
            "otomatic_ai" => $this->getOtomaticAITemplates(),
        ]);
    }

    public function store()
    {
        $this->verifyNonce();

        $validated = $this->validate([
            "name" => ["string", "required"],
            "project" => ["required", "integer"],
        ]);

        $project = Project::find($this->input("project"));

        $payload = [
            "modules" => $project->modules,
        ];

        Arr::forget($payload, 'persona_id');
        Arr::forget($payload, "modules.text.writing_style");
        Arr::forget($payload, "modules.autopilot");
        Arr::forget($payload, "modules.wordpress.post_type");
        Arr::forget($payload, "modules.wordpress.author_id");

        $template = new Template([
            "name" => $validated["name"],
            "plugin_version" => OTOMATIC_AI_VERSION,
            "payload" => $payload,
        ]);

        if ($template->save()) {
            $this->emptyResponse();
        }

        $this->response(["message" => "An error occurred", "error" => "Unable to create the template."], 503);
    }

    public function destroy()
    {
        $this->verifyNonce();

        $this->validate([
            "template" => ["required", "integer"],
        ]);

        $template = Template::find($this->input('template'));
        if ($template) {
            $template->delete();
        }

        $this->emptyResponse();
    }

    public function export()
    {
        $this->verifyNonce();

        $this->validate([
            "template" => ["required", "integer"],
        ]);

        $template = Template::find($this->input('template'));
        if ($template) {

            $payload = $template->payload;

            Arr::forget($payload, 'persona_id');
            Arr::forget($payload, "modules.autopilot");
            Arr::forget($payload, "modules.wordpress");
            Arr::forget($payload, "modules.video");
            Arr::forget($payload, "modules.facebook");
            Arr::forget($payload, "modules.twitter");
            Arr::forget($payload, "modules.instagram");
            Arr::forget($payload, "modules.tiktok");

            $data = [
                "name" => $template->name,
                "plugin_version" => $template->plugin_version,
                "payload" => $payload,
            ];

            $data = json_encode($data);
            $data = base64_encode($data);

            $this->response($data);
        }

        $this->emptyResponse();
    }

    public function import()
    {
        $this->verifyNonce();

        $this->validate([
            "import_data" => ["required", "string"],
        ]);

        $data = $this->input("import_data");
        $data = base64_decode($data);
        $data = json_decode($data, true);

        if (!isset($data["name"])) {
            $this->response(["message" => "The import code is invalid", "errors" => ['import_data' => ["The import code is invalid"]]], 422);
        }


        if (version_compare(Arr::get($data, "plugin_version", "0.0.0"), '4.0.0', '<')) {
            $this->response(["message" => "The template was made with too low a version", "errors" => ['import_data' => ["The template was made with too low a version"]]], 422);
        }

        $payload = Arr::get($data, "payload", []);
        Arr::forget($payload, 'persona_id');
        Arr::forget($payload, "modules.autopilot");
        Arr::forget($payload, "modules.wordpress");
        Arr::forget($payload, "modules.video");
        Arr::forget($payload, "modules.facebook");
        Arr::forget($payload, "modules.twitter");
        Arr::forget($payload, "modules.instagram");
        Arr::forget($payload, "modules.tiktok");

        $template = new Template([
            "name" => Arr::get($data, "name", "new template"),
            "plugin_version" => Arr::get($data, "plugin_version", OTOMATIC_AI_VERSION),
            "payload" => $payload,
        ]);

        $template->save();

        $this->emptyResponse();
    }

    private function getOtomaticAITemplates()
    {
        $optionName = "otomatic_ai_templates";

        $output = [];

        if (function_exists('get_option')) {
            $cache = get_option($optionName, null);

            // fetch or refresh cache
            if (
                empty($cache) ||
                Carbon::createFromFormat(
                    "Y-m-d H:i:s",
                    Arr::get(
                        $cache,
                        "cached_at",
                        Carbon::now()->subMonth()
                    )
                )
                ->lt(
                    Carbon::now()->subMinutes(5)
                )
            ) {
                try {
                    $api = new Client;
                    $output = $api->templates();

                    $cache = [
                        "data" => $output,
                        "cached_at" => Carbon::now()->format("Y-m-d H:i:s"),
                    ];

                    if (function_exists('update_option')) {
                        update_option($optionName, $cache, false);
                    }
                } catch (Exception $e) {
                }
            } else {
                $output = Arr::get($cache, "data", $output);
            }
        }

        return $output;
    }

    private function getLocalTemplates()
    {
        return Template::all()->filter(function (Template $template) {
            return version_compare($template->plugin_version, '4.0.0', '>=');
        })->values();
    }
}
