<?php

namespace OtomaticAi\Jobs;

use DOMDocument;
use DOMXPath;
use Exception;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Models\SingleModeGeneration;
use OtomaticAi\Modules\ProcessCustomFieldsModule;
use OtomaticAi\Modules\ProcessTextModule;
use OtomaticAi\Modules\ProcessImageModule;
use OtomaticAi\Modules\ProcessYoutubeModule;
use OtomaticAi\Modules\ProcessTwitterModule;
use OtomaticAi\Modules\ProcessInstagramModule;
use OtomaticAi\Modules\ProcessTiktokModule;
use OtomaticAi\Modules\ProcessFacebookModule;
use OtomaticAi\Utils\Auth;
use OtomaticAi\Utils\Settings;
use OtomaticAi\Utils\Support;
use OtomaticAi\Vendors\Illuminate\Database\Capsule\Manager as Database;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Carbon;
use Throwable;

class GenerateSingleModeJob extends Job
{
    private SingleModeGeneration $generation;
    private float $startTime;
    private float $endTime;
    private DOMDocument $html;

    /**
     * Create a new job instance.
     */
    public function __construct(SingleModeGeneration $generation)
    {
        $this->generation = $generation;
        $this->generation->load('persona');
        $this->html = new DOMDocument('1.0', 'UTF-8');
    }

    /**
     * Execute the job.
     */
    public function handle()
    {
        global $wpdb;

        try {
            // start timer
            $this->startTimer();

            // verify that the generation can be publish
            if (!$this->canPublish()) {
                return;
            }

            // clear generation logs
            $this->generation->clearLogs();

            // log the start of the job
            $this->generation->addLog("Generation started.");

            // set the generation status to pending
            $this->generation->status = "pending";
            $this->generation->generated_at = Carbon::now();
            $this->generation->save();

            // set the time limit to 1200 seconds
            set_time_limit(1200);
            Database::update("SET SESSION wait_timeout=1200;");
            Database::update("SET SESSION interactive_timeout=1200;");
            $wpdb->query("SET SESSION wait_timeout=1200;");
            $wpdb->query("SET SESSION interactive_timeout=1200;");

            // refresh auth
            Auth::refreshDomain();

            // save start event
            $this->collectEvent("generate-single-mode-start", [
                "title" => $this->generation->title,
                "type" => $this->generation->type,
                "model" => Arr::get($this->generation->modules, "models.text", 'gpt-4o-mini'),
            ]);

            // set the current user
            if (!empty(Arr::get($this->generation->modules, 'wordpress.author_id'))) {
                wp_set_current_user(Arr::get($this->generation->modules, 'wordpress.author_id'));
            } else if (!empty($first = Arr::get(get_users(["fields" => 'ID']), 0))) {
                wp_set_current_user($first);
            }

            // instantiate modules jobs
            $jobs = [
                new ProcessTextModule($this->generation, $this->html),
                new ProcessImageModule($this->generation, $this->html),
                new ProcessYoutubeModule($this->generation, $this->html),
                new ProcessInstagramModule($this->generation, $this->html),
                new ProcessTwitterModule($this->generation, $this->html),
                new ProcessTiktokModule($this->generation, $this->html),
                new ProcessFacebookModule($this->generation, $this->html),
                // new ProcessCustomFieldsModule($this->generation, $this->blocks),
            ];

            // run modules jobs
            foreach ($jobs as $job) {
                $job->handle();
            }

            // remove remaining placeholders
            $xpath = new DOMXPath($this->html);
            $nodes = $xpath->query("//placeholder");
            foreach ($nodes as $node) {
                $node->parentNode->removeChild($node);
            }

            // make content
            $this->generation->content = trim(preg_replace("/^<body>|<\/body>$/", "", $this->html->saveHTML($this->html->getElementsByTagName("body")->item(0))));
            $this->generation->content = wp_encode_emoji($this->generation->content);

            // quit if no content generated
            if (empty($this->generation->content)) {
                return $this->fail(new Exception("No content generated"));
            }

            $this->generation->status = "success";
            $this->generation->save();

            // end timer
            $this->stopTimer();

            $this->generation->addLog("Generation published successfully in " . round($this->getExecutionTime(), 2) . " seconds.", "success");

            // save end event
            $this->collectEvent("generate-single-mode-end", [
                "title" => $this->generation->title,
                "type" => $this->generation->type,
                "model" => Arr::get($this->generation->modules, "models.text", 'gpt-4o-mini'),
            ]);

            // refresh auth
            Auth::refreshDomain();
        } catch (Throwable $e) {
            return $this->fail($e);
        }
    }

    private function canPublish()
    {
        // generation not started
        if ($this->generation->status !== "idle")
            return false;

        // premium or trial
        if (!Auth::isPremium() && !Auth::isTrial()) {
            return false;
        }

        // validate openai signature
        try {
            $response = Auth::validateOpenAISignature(Settings::get("api.openai.api_key"));
            if (!Arr::get($response, "success", false)) {
                return false;
            }
        } catch (Exception $e) {
            return false;
        }

        return true;
    }

    private function fail(Throwable $error)
    {
        $this->generation->addLog("Generation failed. " . $error->getMessage(), "error");

        // save end event
        $this->collectEvent("generate-single-mode-failed", [
            "title" => $this->generation->title,
            "type" => $this->generation->type,
            "model" => Arr::get($this->generation->modules, "models.text", 'gpt-4o-mini'),
            "message" => $error->getMessage(),
        ]);

        $this->generation->status = "failed";
        $this->generation->save();
    }

    private function collectEvent(string $key, array $payload = [])
    {
        $payload["time"] = $this->getExecutionTime();
        $api = new Client;
        $api->collectEvent($key, $payload);
    }

    private function startTimer()
    {
        if (empty($this->startTime)) {
            $this->startTime = microtime(true);
        }
    }

    private function stopTimer()
    {
        if (empty($this->startTime)) {
            return 0;
        }

        if (empty($this->endTime)) {
            $this->endTime = microtime(true);
        }

        // calculate execution time
        return $this->endTime - $this->startTime;
    }

    private function getExecutionTime()
    {
        if (!empty($this->startTime)) {
            $endTime = microtime(true);
            if (!empty($this->endTime)) {
                $endTime = $this->endTime;
            }

            // calculate execution time
            return $endTime - $this->startTime;
        }

        return 0;
    }
}
