<?php

namespace OtomaticAi\Models;

use OtomaticAi\Casts\Language;
use OtomaticAi\Vendors\Illuminate\Database\Eloquent\Model;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;
use OtomaticAi\Vendors\Illuminate\Database\Eloquent\Builder;

class Project extends Model
{
    protected $fillable = ["name", "language", "type", "enabled", "input", "modules", "planning"];

    /**
     * The accessors to append to the model's array form.
     *
     * @var array
     */
    protected $appends = [
        'is_complete',
    ];

    protected $casts = [
        'language' => Language::class,
        'enabled' => 'boolean',
        'input' => 'array',
        'modules' => 'array',
        'planning' => 'array',
        'metrics' => 'array',
    ];

    /**
     * Create a new Eloquent model instance.
     *
     * @param  array  $attributes
     * @return void
     */
    public function __construct(array $attributes = [])
    {
        global $wpdb;

        $this->table = $wpdb->prefix . 'otomatic_ai_projects';
        parent::__construct($attributes);
    }

    /**
     * Get the publications for the project.
     */
    public function publications()
    {
        return $this->hasMany(Publication::class);
    }

    /**
     * Get the persona that owns the project.
     */
    public function persona()
    {
        return $this->belongsTo(Persona::class);
    }

    /**
     * Indicate if the project has publish all his publication
     */
    public function getIsCompleteAttribute()
    {
        return Arr::get($this->metrics, 'idle', 0) + Arr::get($this->metrics, 'pending', 0) === 0;
    }

    public function getModulesAttribute()
    {
        $oldModules = json_decode($this->attributes["modules"], true);
        $newModules = array_replace_recursive([
            "text" => [
                "enabled" => true,
                "model" => "gpt-3.5-turbo",
                "buyer_persona" => [
                    "description" => null
                ],
                "writing_style" => null,
            ],
            "blocks" => [],
            "image" => [
                "service" => "dall_e",
                "custom_instructions" => "",
                "thumbnail" => [
                    "enabled" => false,
                ],
                "settings" => [
                    "stable_diffusion" => [
                        "style_preset" => "",
                        "model" => "stable_image_core"
                    ],
                    "dall_e" => [
                        "model" => "dall-e-3",
                        "quality" => "standard",
                    ],
                    "ideogram" => [
                        "model" => "ideogram_v_2"
                    ],
                    "flux" => [
                        "model" => "flux_1_dev"
                    ],
                    "unsplash" => [],
                ],
            ],
            "wordpress" => [
                "post_type" => "post",
                "author_id" => null,
                "parent_page_id" => null,
                "categories" => [],
                "tags" => [
                    "automatic" => [
                        "enabled" => false,
                    ],
                    "custom" => [],
                ],
                "template" => null,
                "status" => "publish",
                "yoast_seo" => [
                    "title" => [
                        "enabled" => false,
                        "emojis" => [
                            "enabled" => false
                        ]
                    ],
                    "description" => [
                        "enabled" => false,
                    ],
                ],
                "rank_math" => [
                    "title" => [
                        "enabled" => false,
                        "emojis" => [
                            "enabled" => false
                        ]
                    ],
                    "description" => [
                        "enabled" => false,
                    ],
                ],
                "seopress" => [
                    "title" => [
                        "enabled" => false,
                        "emojis" => [
                            "enabled" => false
                        ]
                    ],
                    "description" => [
                        "enabled" => false,
                    ],
                ],
                "custom_fields" => []
            ],
            "autopilot" => [
                "query" => null,
                "source_language" => null,
                "planning" => [
                    "per_day" => 1,
                    "start_time" => [
                        "hours" => 0,
                        "minutes" => 0,
                    ],
                    "end_time" => [
                        "hours" => 23,
                        "minutes" => 59,
                    ],
                ]
            ]
        ], $oldModules);

        // set ai model to default if no longer available
        if (!in_array(Arr::get($newModules, "text.model"), ["gpt-4-turbo", "gpt-4o", "gpt-4o-mini", "mistral-large", "groq-llama3-70b", "claude-3-5-sonnet"])) {
            Arr::set($newModules, "text.model", "gpt-4o-mini");
        }

        // set ai model to default if not available for scrapping
        if (in_array($this->type, ["news", "news-now", "rss", "rss-now", "url", "sitemap"]) && !in_array(Arr::get($newModules, "text.model"), ["gpt-4-turbo", "gpt-4o", "gpt-4o-mini", "mistral-large", "claude-3-5-sonnet"])) {
            Arr::set($newModules, "text.model", "gpt-4o-mini");
        }

        // add automatic categories
        if (Arr::isList($oldModules["wordpress"]["categories"])) {
            $newModules["wordpress"]["categories"] = [
                "automatic" => [
                    "enabled" => false,
                ],
                "custom" => $oldModules["wordpress"]["categories"]
            ];
        }

        // set post type to page if project type is siloing
        if ($this->type === "siloing") {
            Arr::set($newModules, "wordpress.post_type", "page");
        }

        // try to set the wordpress.author_id if is not set and persona is set
        if ($this->persona_id !== null && empty(Arr::get($newModules, "wordpress.author_id"))) {
            $persona = $this->persona;
            if (!empty($persona)) {
                $author = $persona->user;
                if (!empty($author)) {
                    Arr::set($newModules, "wordpress.author_id", $author->ID);
                }
            }
        }

        // change stable diffusion sd3 model to sd3-large if not set
        if (Arr::get($newModules, "image.settings.stable_diffusion.model") === "stable_image_sd3") {
            Arr::set($newModules, "image.settings.stable_diffusion.model", "stable_image_sd3_large");
        }
        if (Arr::get($newModules, "image.settings.stable_diffusion.model") === "stable_image_sd3_turbo") {
            Arr::set($newModules, "image.settings.stable_diffusion.model", "stable_image_sd3_large_turbo");
        }

        // make default blocks module if not exist
        if (empty(Arr::get($newModules, "blocks", []))) {
            $blocks = [];
            $externalLinksEnabled = Arr::get($oldModules, "text.external_links.enabled", false);

            // brief
            if (Arr::get($oldModules, "text.options.brief.enabled", false)) {
                $blocks[] = $this->makeDefaultTextBlock("brief", true, false);
            }

            // youtube
            if (Arr::get($oldModules, "video.enabled", false)) {
                $blocks[] = ["block" => "youtube"];
            }

            // introduction
            if (Arr::get($oldModules, "text.options.introduction.enabled", false)) {
                $blocks[] = $this->makeDefaultTextBlock("introduction", true, false);
            }

            // content image 1
            if (Arr::get($oldModules, "image.content.enabled", false) && Arr::get($oldModules, "image.content.length", 0) > 0) {
                $blocks[] = ["block" => "image"];
            }

            // text content
            $blocks[] = $this->makeDefaultTextBlock("content", true, $externalLinksEnabled);

            // content image 1
            if (Arr::get($oldModules, "image.content.enabled", false) && Arr::get($oldModules, "image.content.length", 0) > 1) {
                $blocks[] = ["block" => "image"];
            }

            // table
            if (Arr::get($oldModules, "text.options.table.enabled", false)) {
                $blocks[] = $this->makeDefaultTextBlock("table", true, false);
            }

            // list
            if (Arr::get($oldModules, "text.options.list.enabled", false)) {
                $blocks[] = $this->makeDefaultTextBlock("list", true, false);
            }

            // tiktok
            if (Arr::get($oldModules, "tiktok.enabled", false)) {
                $blocks[] = ["block" => "tiktok"];
            }

            // instagram
            if (Arr::get($oldModules, "instagram.enabled", false)) {
                $blocks[] = ["block" => "instagram"];
            }

            // twitter
            if (Arr::get($oldModules, "twitter.enabled", false)) {
                $blocks[] = ["block" => "twitter"];
            }

            // facebook
            if (Arr::get($oldModules, "facebook.enabled", false)) {
                $blocks[] = ["block" => "facebook"];
            }

            // faq
            if (Arr::get($oldModules, "text.options.faq.enabled", false)) {
                $blocks[] = $this->makeDefaultTextBlock("faq", true, false);
            }

            Arr::set($newModules, "blocks", $blocks);
        }


        return $newModules;
    }

    /**
     * Scope a query to only include completed publications.
     */
    public function scopeCompleted(Builder $query): void
    {
        $query->whereRaw("JSON_UNQUOTE(JSON_EXTRACT(metrics, '$.idle')) = ?", [0])
            ->whereRaw("JSON_UNQUOTE(JSON_EXTRACT(metrics, '$.pending')) = ?", [0]);
    }

    public function refreshMetrics()
    {
        $this->metrics = $this->publications()->toBase()
            ->selectRaw("count(*) as total")
            ->selectRaw("count(IF(status = 'idle', 1, null)) as idle")
            ->selectRaw("count(IF(status = 'pending', 1, null)) as pending")
            ->selectRaw("count(IF(status = 'success', 1, null)) as success")
            ->selectRaw("count(IF(status = 'failed', 1, null)) as failed")
            ->first();
        $this->save();
    }

    public function getFormattedModules()
    {
        $modules = $this->modules;

        // preset
        if (Arr::get($modules, 'text.template') === 'custom') {
            $preset = Preset::find(Arr::get($modules, 'text.settings.custom.preset'));
            if ($preset)
                Arr::set($modules, 'text.settings.custom.preset', Arr::only($preset->toArray(), ['model', 'messages', 'temperature', 'top_p', 'presence_penalty', 'frequency_penalty']));
            else
                Arr::set($modules, 'text.settings.custom.preset', null);
        }

        return $modules;
    }

    private function makeDefaultTextBlock(string $block, bool $useLiveData = true, bool $externalLinksEnabled = false, bool $internalLinksEnabled = false)
    {
        $prompt = "";

        switch ($block) {
            case "brief":
                $prompt = "écrit un brief pertinent  sous forme de liste au format HTML, très concis, qui apporte un premier coup d'œil au contenu de l'article au lecteur sur le sujet : [TITLE].\n\nle contenu de ce html doit être encadré dans un tableau.\n\nAjoute un titre : EN BREF en [LANGUAGE] (en <p></p>)\n\nMet les mots clés important en <strong>.";
                break;
            case "introduction":
                $prompt = "Ecrit une introduction courte sur le sujet : [TITLE]\n\nne met pas de titre ou de sous-titre.";
                break;
            case "table":
                $prompt = "créer un tableau comparateur au format HTML pertinent avec deux colonnes , très concis sur :\n[TITLE]\n\nAjoute un sous titre en <h2>.\n\nTrouve un axe pertinent pour créer le tableau, ce n'est pas forcément un résumé. La colonne de droite seulement un texte concis. Je veux un vrai tableau mais pas plus de 10 lignes.\n\nNe met pas de guillemets.\n\nN hesites pas à rajouté des colonnes si nécessaire.\n\nNe met pas de titre ou de sous-titre.";
                break;
            case "list":
                $prompt = "créer une liste au format HTML pertinent avec deux colonnes , très concis sur :\n[TITLE].\n\nTrouve un axe pertinent pour créer la liste, ce n'est pas forcément un résumé.\n\nNe genere pas de tableau.";
                break;
            case "content":
                $prompt = "Ecrit un article long sur la sujet : \"[TITLE]\".\n\nInclure des sous-titres qui correspondent à : \"[TITLE]\"  sans aucune numérotation. \nStructure le texte en paragraphes clairs (entre <p> et </p>) et distincts pour une lecture aisée.\n\nStructure le texte avec des sous-titres, en utilisant les balises HTML <h2> et <h3>.\nEcrit un premier paragraphe clair en début d'article qui résume rapidement ce meme article, et des paragraphes sous les <h2>.";
                break;
        }

        return [
            "block" => "text",
            "settings" => [
                "name" => Str::headline($block),
                "use_live_data" => $useLiveData,
                "external_links_enabled" => $externalLinksEnabled,
                "internal_links_enabled" => $internalLinksEnabled,
                "prompt" => $prompt,
            ]
        ];
    }
}
