<?php

namespace OtomaticAi\Models;

use OtomaticAi\Casts\Language;
use OtomaticAi\Utils\Image;
use OtomaticAi\Vendors\Illuminate\Database\Eloquent\Model;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Database\Eloquent\Builder;

class Project extends Model
{
    static $version = 2;

    protected $fillable = ["name", "language", "type", "enabled", "input", "modules", "planning"];

    /**
     * The accessors to append to the model's array form.
     *
     * @var array
     */
    protected $appends = [
        'is_complete',
    ];

    protected $casts = [
        'language' => Language::class,
        'enabled' => 'boolean',
        'input' => 'array',
        'modules' => 'array',
        'planning' => 'array',
        'metrics' => 'array',
    ];

    /**
     * Create a new Eloquent model instance.
     *
     * @param  array  $attributes
     * @return void
     */
    public function __construct(array $attributes = [])
    {
        global $wpdb;

        $this->table = $wpdb->prefix . 'otomatic_ai_projects';
        parent::__construct($attributes);
    }

    /**
     * Get the publications for the project.
     */
    public function publications()
    {
        return $this->hasMany(Publication::class);
    }

    /**
     * Get the persona that owns the project.
     */
    public function persona()
    {
        return $this->belongsTo(Persona::class);
    }

    /**
     * Indicate if the project has publish all his publication
     */
    public function getIsCompleteAttribute()
    {
        return Arr::get($this->metrics, 'idle', 0) + Arr::get($this->metrics, 'pending', 0) === 0;
    }

    public function getModulesAttribute()
    {
        $newModules = self::makeModules(json_decode($this->attributes["modules"], true));

        // set post type to page if project type is siloing
        if ($this->type === "siloing") {
            Arr::set($newModules, "wordpress.post_type", "page");
        }

        // try to set the wordpress.author_id if is not set and persona is set
        if ($this->persona_id !== null && empty(Arr::get($newModules, "wordpress.author_id"))) {
            $persona = $this->persona;
            if (!empty($persona)) {
                $author = $persona->user;
                if (!empty($author)) {
                    Arr::set($newModules, "wordpress.author_id", $author->ID);
                }
            }
        }

        // disable custom links
        Arr::set($newModules, "wordpress.custom_links.enabled", false);

        return $newModules;
    }

    private static function transformOldModules(array $modules)
    {
        if (Arr::get($modules, "_version", 1) < 2) {
            return array_replace_recursive([
                "models" => [
                    "text" => Arr::get($modules, "text.model", "gpt-4o-mini"),
                    "image" => Image::imageProviderAndModelToModel(Arr::get($modules, "image.service"), Arr::get($modules, "image.settings." . Arr::get($modules, "image.service") . ".model")),
                ],
                "content" => [
                    "writing_style" => Arr::get($modules, "text.writing_style"),
                    "sources" => [
                        "enabled" => Arr::get($modules, "text.sources.enabled", false),
                        "no_follow" => Arr::get($modules, "text.sources.no_follow", false),
                        "blank" => Arr::get($modules, "text.sources.blank", false),
                    ],
                    "custom_instructions" => [
                        "images" => Arr::get($modules, "image.custom_instructions"),
                    ],
                    "blocks" => !Arr::get($modules, "automatic_template.enabled", false) ? Arr::get($modules, "blocks", []) : [],
                    "automatic" => [
                        "enabled" => Arr::get($modules, "automatic_template.enabled", false) || empty(Arr::get($modules, "blocks", [])),
                        "images" => [
                            "enabled" => !Arr::get($modules, "automatic_template.disable_image_generation", false),
                        ],
                        "youtube" => [
                            "enabled" => !Arr::get($modules, "automatic_template.disable_youtube_generation", false),
                        ],
                        "embeds" => [
                            "enabled" => !Arr::get($modules, "automatic_template.disable_embeds_generation", false),
                        ],
                        "internal_links" => [
                            "enabled" => !Arr::get($modules, "automatic_template.disable_internal_links", false),
                        ],
                        "external_links" => [
                            "enabled" => !Arr::get($modules, "automatic_template.disable_external_links", false),
                        ],

                    ],
                ],
                "wordpress" => [
                    "thumbnail" => [
                        "enabled" => Arr::get($modules, "image.thumbnail.enabled", false),
                    ],
                    "seo_title" => [
                        "enabled" => Arr::get($modules, "wordpress.yoast_seo.title.enabled", false) || Arr::get($modules, "wordpress.rank_math.title.enabled", false) || Arr::get($modules, "wordpress.seo_title.enabled", false),
                        "emojis" => [
                            "enabled" => Arr::get($modules, "wordpress.yoast_seo.title.emojis.enabled", false) || Arr::get($modules, "wordpress.rank_math.title.emojis.enabled", false) || Arr::get($modules, "wordpress.seo_title.emojis.enabled", false),
                        ]
                    ],
                    "seo_description" => [
                        "enabled" => Arr::get($modules, "wordpress.yoast_seo.description.enabled", false) || Arr::get($modules, "wordpress.rank_math.description.enabled", false) || Arr::get($modules, "wordpress.seo_description.enabled", false),
                    ],
                ],
            ], $modules);
        }

        return $modules;
    }

    public static function makeModules(array $modules)
    {
        $modules = self::transformOldModules($modules);

        $newModules = array_replace_recursive([
            "_version" => static::$version,
            "models" => [
                "text" => "gpt-4o-mini",
                "image" => "flux_1_dev"
            ],
            "content" => [
                "writing_style" => null,
                "sources" => [
                    "enabled" => false,
                    "no_follow" => false,
                    "blank" => false,
                ],
                "custom_instructions" => [
                    "profile" => null,
                    "images" => null,
                ],
                "blocks" => [],
                "automatic" => [
                    "enabled" => true,
                    "text_length" => "medium",
                    "images" => [
                        "enabled" => true,
                        "max_images" => 3
                    ],
                    "youtube" => [
                        "enabled" => true,
                    ],
                    "embeds" => [
                        "enabled" => true,
                    ],
                    "internal_links" => [
                        "enabled" => true,
                    ],
                    "external_links" => [
                        "enabled" => true,
                    ],
                    "custom_links" => [
                        "enabled" => false,
                        "links" => [],
                    ],
                    "lists" => [
                        "enabled" => true,
                    ],
                    "tables" => [
                        "enabled" => true,
                    ],
                    "bold_words" => [
                        "enabled" => true,
                    ],
                    "brands" => [
                        "enabled" => true,
                    ],
                    "emojis" => [
                        "enabled" => false,
                    ],
                    "summary" => [
                        "enabled" => false,
                    ],
                    "faq" => [
                        "enabled" => false,
                    ],
                    "keywords" => [
                        "automatic" => [
                            "enabled" => true,
                        ],
                        "custom" => [],
                    ],
                ],
            ],
            "wordpress" => [
                "thumbnail" => [
                    "enabled" => true,
                ],
                "post_type" => "post",
                "author_id" => null,
                "parent_page_id" => null,
                "categories" => [
                    "automatic" => [
                        "enabled" => false,
                    ],
                    "custom" => [],
                ],
                "tags" => [
                    "automatic" => [
                        "enabled" => false,
                    ],
                    "custom" => [],
                ],
                "template" => null,
                "status" => "publish",
                "rewrite_permalink" => [
                    "enabled" => false,
                ],
                "seo_title" => [
                    "enabled" => false,
                    "emojis" => [
                        "enabled" => false
                    ]
                ],
                "seo_description" => [
                    "enabled" => false,
                ],
                "custom_fields" => []
            ],
            "autopilot" => [
                "query" => null,
                "source_language" => null,
                "planning" => [
                    "per_day" => 1,
                    "start_time" => [
                        "hours" => 0,
                        "minutes" => 0,
                    ],
                    "end_time" => [
                        "hours" => 23,
                        "minutes" => 59,
                    ],
                ]
            ],
        ], $modules);

        // convert old models to new models
        $newModules = self::convertOldTextModels($newModules);

        return $newModules;
    }

    /**
     * Scope a query to only include completed publications.
     */
    public function scopeCompleted(Builder $query): void
    {
        $query->whereRaw("JSON_UNQUOTE(JSON_EXTRACT(metrics, '$.idle')) = ?", [0])
            ->whereRaw("JSON_UNQUOTE(JSON_EXTRACT(metrics, '$.pending')) = ?", [0]);
    }

    public function refreshMetrics()
    {
        $this->metrics = $this->publications()->toBase()
            ->selectRaw("count(*) as total")
            ->selectRaw("count(IF(status = 'idle', 1, null)) as idle")
            ->selectRaw("count(IF(status = 'pending', 1, null)) as pending")
            ->selectRaw("count(IF(status = 'success', 1, null)) as success")
            ->selectRaw("count(IF(status = 'failed', 1, null)) as failed")
            ->first();
        $this->save();
    }

    /**
     * Convert old text models to new text models
     */
    private static function convertOldTextModels(array $modules)
    {
        // set ai model to default if no longer available
        if (!in_array(Arr::get($modules, "models.text"), [
            "gpt-4o-mini",
            "gpt-4o",
            "o1",
            "o1-mini",
            "o3-mini",
            "gpt-4-5",
            "mistral-large",
            "groq-llama3-70b",
            "claude-3-5-haiku",
            "claude-3-5-sonnet",
            "claude-3-7-sonnet",
            "gemini-2-flash-lite",
            "gemini-2-flash",
            "gemini-2-pro",
            "gemini-2-flash-thinking",
            "deepseek-chat",
            "deepseek-reasoner",
            "grok-2"
        ])) {
            Arr::set($modules, "models.text", "gpt-4o-mini");
        }

        return $modules;
    }
}
