<?php

namespace OtomaticAi\Modules;

use Exception;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Content\Section;
use OtomaticAi\Content\Social\Facebook;
use OtomaticAi\Models\Presets\Preset;
use OtomaticAi\Modules\Contracts\Module as ModuleContract;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;

class ProcessFacebookModule extends Module implements ModuleContract
{
    /**
     * Execute the job.
     *
     * @return void
     * @throws Exception
     */
    public function handle(): void
    {
        // verify that the job is runnable
        if (!$this->isRunnable()) {
            return;
        }

        // log the start of the job
        $this->publication->addLog("Facebook module started.");

        // get length
        $length = $this->getLength();

        // create facebook placeholders
        $placeholders = $this->createElementPlaceholders("facebook");

        try {
            // make the search terms
            $search = $this->makeSearch($this->getTitle());

            // call the api
            $api = new Client;
            $result = $api->facebook($search, $this->getLanguage()->key);

            // get the organic results
            $posts = is_array($result) ? $result : [];

            // randomize posts
            $posts = Arr::shuffle($posts);

            // transform link into displayable link
            $posts = array_map(function ($item) {
                $parse = parse_url($item["link"]);
                if (!empty($parse["path"]) && !empty($parse["scheme"]) && !empty($parse["host"])) {
                    preg_match_all('/\/(.*?)\/posts\/.*?\/(.*?)\//', $parse["path"], $matches, PREG_SET_ORDER);
                    if (!empty($matches) && count($matches[0]) === 3) {
                        $item["link"] = $parse["scheme"] . "://" . $parse["host"] . '/' . $matches[0][1] . '/posts/' . $matches[0][2];
                        return $item;
                    }
                }

                return null;
            }, $posts);

            // remove empty values
            $posts = array_values(array_filter($posts));

            if (!empty($posts)) {

                // get x random posts
                $posts = Arr::random($posts, min(count($posts), $length));
                $posts = array_map(function ($post) {
                    return new Facebook($post["link"]);
                }, $posts);

                // replace facebook placeholders with the generated posts
                $this->replaceElementPlaceholders("facebook", function ($attributes) use ($posts) {
                    $index = Arr::get($attributes, "index");
                    if ($index !== null && isset($posts[$index])) {
                        $post = $posts[$index];

                        $el = $post->toHtmlElement($this->html);
                        if ($el) {
                            return $el;
                        }
                    }
                });

                // log the end of the job
                $this->publication->addLog("Facebook module completed successfully.", "success");
            } else {
                $this->publication->addLog("Facebook module failed. No posts found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Facebook module failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Make the Facebook search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("get_main_keyword");

        $response = $preset->process([
            "language" => $this->getLanguage()->value,
            "title" => $this->getTitle(),
            "request" => $this->getTitle(),
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Get the number of Facebook elements
     *
     * @return integer
     */
    private function getLength(): int
    {
        return $this->countElements("facebook");;
    }

    /**
     * Determine if the module is runnable
     *
     * @return boolean
     */
    public function isRunnable(): bool
    {
        // must have facebook elements
        return $this->getLength() > 0;
    }
}
