<?php

namespace OtomaticAi\Modules;

use Exception;
use OtomaticAi\Api\OpenAi\PoolClient as OpenAiPoolClient;
use OtomaticAi\Api\StabilityAi\PoolClient as StabilityAiPoolClient;
use OtomaticAi\Api\Ideogram\PoolClient as IdeogramPoolClient;
use OtomaticAi\Api\Flux\PoolClient as FluxPoolClient;
use OtomaticAi\Api\Unsplash\Client as UnsplashClient;
use OtomaticAi\Api\Pexels\Client as PexelsClient;
use OtomaticAi\Api\Pixabay\Client as PixabayClient;
use OtomaticAi\Api\GoogleImage\Client as GoogleImageClient;
use OtomaticAi\Api\BingImage\Client as BingImageClient;
use OtomaticAi\Content\Image\BingImage;
use OtomaticAi\Content\Image\DallE;
use OtomaticAi\Content\Image\Flux;
use OtomaticAi\Content\Image\GoogleImage;
use OtomaticAi\Content\Image\Pexels;
use OtomaticAi\Content\Image\Pixabay;
use OtomaticAi\Content\Image\StableDiffusion;
use OtomaticAi\Content\Image\Ideogram;
use OtomaticAi\Content\Image\Unsplash;
use OtomaticAi\Content\Section;
use OtomaticAi\Models\Preset;
use OtomaticAi\Models\Publication;
use OtomaticAi\Modules\Contracts\Module as ModuleContract;
use OtomaticAi\Utils\Language;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;

class ProcessImageModule implements ModuleContract
{
    /**
     * The publication
     *
     * @var Publication
     */
    private Publication $publication;

    /**
     * Array of used image urls
     *
     * @var array
     */
    private array $usedImages = [];

    /**
     * Create a new job instance.
     *
     * @param Publication $publication
     */
    public function __construct(Publication $publication)
    {
        $this->publication = $publication;
    }

    /**
     * Execute the job.
     *
     * @return void
     * @throws Exception
     */
    public function handle(): void
    {
        // verify that the job is runnable
        if (!self::isRunnable($this->publication)) {
            return;
        }

        // log the start of the job
        $this->publication->addLog("Image module started.");

        // determine length
        $length = self::getLength($this->publication);
        if ($length === 0) {
            return;
        }

        // generate images
        $images = [];
        switch (Arr::get($this->publication->project->modules, "image.service")) {
            case "stable_diffusion":
                $images = $this->getStableDiffusionImage($this->publication->generation_title, $length);
                break;
            case "ideogram":
                $images = $this->getIdeogramImage($this->publication->generation_title, $length);
                break;
            case "flux":
                $images = $this->getFluxImage($this->publication->generation_title, $length);
                break;
            case "dall_e":
                $images = $this->getDallEImage($this->publication->generation_title, $length);
                break;
            case "unsplash":
                $images = $this->getUnsplashImage($this->publication->generation_title, $length);
                break;
            case "pexels":
                $images = $this->getPexelsImage($this->publication->generation_title, $length);
                break;
            case "pixabay":
                $images = $this->getPixabayImage($this->publication->generation_title, $length);
                break;
            case "google_image":
                $images = $this->getGoogleImage($this->publication->generation_title, $length);
                break;
            case "bing_image":
                $images = $this->getBingImage($this->publication->generation_title, $length);
                break;
        }

        $initialLength = $length;
        $successfulGenerations = 0;

        // add images to sections
        if (!empty($images)) {
            $successfulGenerations = count($images);

            // thumbnail image
            if (Arr::get($this->publication->project->modules, "image.thumbnail.enabled", false)) {
                $extras = $this->publication->extras;
                $extras["thumbnail"] = array_shift($images)->toArray();
                $this->publication->extras = $extras;
            }

            if (!empty($images)) {

                // get publication sections
                $sections = $this->publication->sections;

                // content images
                $index = 0;
                foreach (Arr::get($this->publication->project->modules, 'blocks', []) as $blockIndex => $block) {
                    if ($block["block"] === "image" && isset($images[$index])) {
                        $section = new Section();
                        $section->addElement($images[$index]);
                        $sections->put($blockIndex, $section);
                        $index++;
                    }
                }

                // update publication sections
                $this->publication->sections = $sections;
            }
        }

        // log the end of the job
        if ($successfulGenerations === $initialLength) {
            $this->publication->addLog("Images module completed successfully.", "success");
        } else {
            $this->publication->addLog("Images module completed with errors. " . $successfulGenerations . " on " . $initialLength . " images generated.", "warning");
        }
    }

    /**
     * Generate a stable diffusion image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return StableDiffusion|StableDiffusion[]|null
     */
    private function getStableDiffusionImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Stable Diffusion started.");

        try {
            // make the prompt for stable diffusion
            $prompt = $this->makeStableDiffusionPrompt($search);

            // make the pool settings
            $payloads = [];
            for ($i = 0; $i < $length; $i++) {

                if (Arr::get($this->publication->project->modules, "image.settings.stable_diffusion.model") === "sdxl") {
                    // sdxl model
                    $payload = [
                        "text_prompts" => [
                            [
                                "text" => $prompt,
                                "weight" => 1
                            ]
                        ]
                    ];
                } else {
                    // others models
                    $payload = [
                        "prompt" => $prompt,
                    ];

                    switch (Arr::get($this->publication->project->modules, "image.settings.stable_diffusion.model")) {
                        case "stable_image_core":
                            $payload["model"] = "core";
                            break;
                        case "stable_image_sd3_medium":
                            $payload["model"] = "sd3-medium";
                            break;
                        case "stable_image_sd3_turbo":
                        case "stable_image_sd3_large_turbo":
                            $payload["model"] = "sd3-large-turbo";
                            break;
                        case "stable_image_sd3":
                        case "stable_image_sd3_large":
                            $payload["model"] = "sd3-large";
                            break;
                        case "stable_image_sd3_5_large":
                            $payload["model"] = "sd3.5-large";
                            break;
                        case "stable_image_sd3_5_large_turbo":
                            $payload["model"] = "sd3.5-large-turbo";
                            break;
                        case "stable_image_ultra":
                            $payload["model"] = "ultra";
                            break;
                        default:
                            $payload["model"] = "core";
                    }
                }
                $payloads[] = $payload;
            }

            // call the stability ai api
            $api = new StabilityAiPoolClient();
            if (Arr::get($this->publication->project->modules, "image.settings.stable_diffusion.model") === "sdxl") {
                // sdxl model
                $result = [];
                foreach ($api->textToImage($payloads) as $artifact) {
                    if (Arr::get($artifact, "artifacts.0.finishReason") === "SUCCESS") {
                        $result[] = [
                            "image" => Arr::get($artifact, "artifacts.0.base64")
                        ];
                    }
                }
            } else {
                // others models
                $result = $api->stableImageGenerate($payloads);
            }

            // make the Stable Diffusion Images
            $images = [];
            foreach ($result as $res) {

                if (!empty(Arr::get($res, "image", null))) {
                    // description
                    $description = $this->makeImageDescription($search);

                    $images[] = new StableDiffusion(Arr::get($res, "image"), $search, $description);
                }
            }
            return $images;
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate a ideogram image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return Ideogram|Ideogram[]|null
     */
    private function getIdeogramImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Ideogram started.");

        try {
            // make the prompt for ideogram
            $prompt = $this->makeIdeogramPrompt($search);

            // make the pool settings
            $payloads = [];
            for ($i = 0; $i < $length; $i++) {
                $payload = [
                    "image_request" => [
                        "prompt" => $prompt,
                    ]
                ];
                switch (Arr::get($this->publication->project->modules, "image.settings.ideogram.model")) {
                    case "ideogram_v_1":
                        $payload["image_request"]["model"] = "V_1";
                        break;
                    case "ideogram_v_1_turbo":
                        $payload["image_request"]["model"] = "V_1_TURBO";
                        break;
                    case "ideogram_v_2":
                        $payload["image_request"]["model"] = "V_2";
                        break;
                    case "ideogram_v_2_turbo":
                        $payload["image_request"]["model"] = "V_2_TURBO";
                        break;
                    default:
                        $payload["image_request"]["model"] = "V_2";
                }

                $payloads[] = $payload;
            }

            // call the ideogram api
            $api = new IdeogramPoolClient();
            $result = $api->generate($payloads);

            // make the Ideogram Images
            $images = [];
            foreach ($result as $res) {

                if (!empty(Arr::get($res, "data.0.url", null))) {
                    // description
                    $description = $this->makeImageDescription($search);

                    $images[] = new Ideogram(Arr::get($res, "data.0.url"), $search, $description);
                }
            }
            return $images;
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate a Flux image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return Flux|Flux[]|null
     */
    private function getFluxImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Flux started.");

        try {
            // make the prompt for flux
            $prompt = $this->makeFluxPrompt($search);

            // make the pool settings
            $payloads = [];
            for ($i = 0; $i < $length; $i++) {


                // others models
                $payload = [
                    "prompt" => $prompt,
                ];

                switch (Arr::get($this->publication->project->modules, "image.settings.flux.model")) {
                    case "flux_1_dev":
                        $payload["model"] = "flux-dev";
                        break;
                    case "flux_1_pro":
                        $payload["model"] = "flux-pro";
                        break;
                    case "flux_1_1_pro":
                        $payload["model"] = "flux-pro-1.1";
                        break;
                    default:
                        $payload["model"] = "flux-dev";
                }

                $payloads[] = $payload;
            }

            $api = new FluxPoolClient();

            // create the tasks
            $responses = $api->createTasks($payloads);

            // get the tasks ids
            $ids = array_map(function ($response) {
                return $response["id"];
            }, $responses);

            // get the tasks results
            $results = [];
            while (true) {
                sleep(3);

                $results = $api->getResults($ids);

                $again = false;
                foreach ($results as $result) {
                    if (Arr::get($result, "status") === "Pending") {
                        $again  = true;
                        break;
                    }
                }

                if (!$again) {
                    break;
                }
            }

            // make the Flux Images
            $images = [];
            foreach ($results as $res) {

                if (!empty(Arr::get($res, "result.sample", null))) {
                    // description
                    $description = $this->makeImageDescription($search);

                    $images[] = new Flux(Arr::get($res, "result.sample"), $search, $description);
                }
            }

            return $images;
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate a dall-e image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return DallE|DallE[]|null
     */
    private function getDallEImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Dall-E started.");

        try {
            // make the prompt for dall e
            $prompt = $this->makeDallEPrompt($search);

            // make the pool settings
            $settings = [
                "prompts" => [],
                "settings" => [],
            ];
            for ($i = 0; $i < $length; $i++) {
                $settings["prompts"][] = $prompt;
                $model = Arr::get($this->publication->project->modules, "image.settings.dall_e.model", "dall-e-2");
                $settings["settings"][] = [
                    "model" => $model,
                    "quality" => Arr::get($this->publication->project->modules, "image.settings.dall_e.quality", "standard"),
                    "size" => $model === "dall-e-3" ? "1792x1024" : "1024x1024",
                ];
            }

            // call the openai api
            $api = new OpenAiPoolClient();
            $result = $api->image($settings["prompts"], $settings["settings"]);

            // make the DallE Images
            $images = [];
            foreach ($result as $res) {

                // description
                $description = $this->makeImageDescription($search);

                $images[] = new DallE(Arr::get($res, "data.0.b64_json"), $search, $description);
            }
            return $images;
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate an Unsplash image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return Unsplash|Unsplash[]|null
     */
    private function getUnsplashImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Unsplash started.");

        try {

            // make the search for unsplash
            $search = $this->makeUnsplashSearch($search);

            // call unsplash api
            $api = new UnsplashClient;
            $response = $api->searchPhotos([
                "query" => $search,
                "per_page" => 20,
            ]);

            // get a random image
            $result = Arr::get($response, "results", []);

            // filter to remove used images
            $result = array_filter($result, function ($image) {
                return Arr::has($image, "urls.regular") && !in_array($image["urls"]["regular"], $this->usedImages);
            });
            $result = array_values($result);

            if (!empty($result)) {

                $images = [];
                $result = Arr::random($result, min(count($result), $length));

                foreach ($result as $res) {

                    // description
                    $description = $this->makeImageDescription($search);

                    // add image to used images
                    $this->usedImages[] = $res["urls"]["regular"];

                    $images[] = new Unsplash($res["urls"]["regular"], $search, $description);
                }

                return $images;
            } else {
                $this->publication->addLog("Image generation failed. No images found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate an Pexels image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return Pexel|Pexel[]|null
     */
    private function getPexelsImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Pexels started.");

        try {

            // make the search for pexels
            $search = $this->makePexelsSearch($search);

            // call pexels api
            $api = new PexelsClient();
            $response = $api->searchPhotos([
                "query" => $search,
                "per_page" => 20,
            ]);

            // get a random image
            $result = Arr::get($response, "photos", []);

            // filter to remove used images
            $result = array_filter($result, function ($image) {
                return !in_array($image["src"]["landscape"], $this->usedImages);
            });
            $result = array_values($result);

            if (!empty($result)) {

                $images = [];
                $result = Arr::random($result, min(count($result), $length));
                foreach ($result as $res) {

                    // description
                    $description = $this->makeImageDescription($search);

                    // add image to used images
                    $this->usedImages[] = $res["src"]["landscape"];

                    $images[] = new Pexels($res["src"]["landscape"], $search, $description);
                }

                return $images;
            } else {
                $this->publication->addLog("Image generation failed. No images found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate an Pixabay image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return Pixabay|Pixabay[]|null
     */
    private function getPixabayImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Pixabay started.");

        try {

            // make the search for pixabay
            $search = $this->makePixabaySearch($search);

            // call pixabay api
            $api = new PixabayClient();
            $response = $api->images([
                "q" => $search,
            ]);

            // get a random image
            $result = Arr::get($response, "hits", []);

            // filter to remove used images
            $result = array_filter($result, function ($image) {
                return !in_array($image["largeImageURL"], $this->usedImages);
            });
            $result = array_values($result);

            if (!empty($result)) {

                $images = [];
                $result = Arr::random($result, min(count($result), $length));
                foreach ($result as $res) {

                    // description
                    $description = $this->makeImageDescription($search);

                    // add image to used images
                    $this->usedImages[] = $res["largeImageURL"];

                    $images[] = new Pixabay($res["largeImageURL"], $search, $description);
                }

                return $images;
            } else {
                $this->publication->addLog("Image generation failed. No images found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate an Google image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return GoogleImage|GoogleImage[]|null
     */
    private function getGoogleImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Google Image started.");

        try {

            // make the search for google image
            $search = $this->makeGoogleImageSearch($search);

            // call google image api
            $api = new GoogleImageClient();
            $result = $api->images([
                "q" => $search,
            ]);

            // filter to remove used images
            $result = array_filter($result, function ($url) {
                return !in_array($url, $this->usedImages);
            });
            $result = array_values($result);

            // get a random image
            if (!empty($result)) {

                $images = [];
                $result = Arr::random($result, min(count($result), $length));
                foreach ($result as $res) {

                    // description
                    $description = $this->makeImageDescription($search);

                    // add image to used images
                    $this->usedImages[] = $res;

                    $images[] = new GoogleImage($res, $search, $description);
                }

                return $images;
            } else {
                $this->publication->addLog("Image generation failed. No images found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate an Bing image for the provided search
     *
     * @param string $search
     * @param int $length
     * @return BingImage|BingImage[]|null
     */
    private function getBingImage(string $search, int $length = 1)
    {
        // log the image generation
        $this->publication->addLog("Image generation with Bing Image started.");

        try {

            // make the search for bing image
            $search = $this->makeBingImageSearch($search);

            // call google image api
            $api = new BingImageClient();
            $result = $api->images([
                "q" => $search,
            ]);

            // filter to remove used images
            $result = array_filter($result, function ($url) {
                return !in_array($url, $this->usedImages);
            });
            $result = array_values($result);

            // get a random image
            if (!empty($result)) {

                $images = [];
                $result = Arr::random($result, min(count($result), $length));
                foreach ($result as $res) {

                    // description
                    $description = $this->makeImageDescription($search);

                    // add image to used images
                    $this->usedImages[] = $res;

                    $images[] = new BingImage($res, $search, $description);
                }

                return $images;
            } else {
                $this->publication->addLog("Image generation failed. No images found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Image generation failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Generate the stable diffusion prompt for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeStableDiffusionPrompt(string $search): string
    {
        $search = Str::lower($search);
        $customInstructions = $this->getCustomInstructions();

        $preset = Preset::findFromAPI("generate_stable_diffusion_prompt");

        $response = $preset->process([
            "request" => $search,
            "has_custom_instructions" => !empty($customInstructions),
            "custom_instructions" => $customInstructions,
        ]);

        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        if (!empty($content)) {
            return $content;
        }

        return $search;
    }

    /**
     * Generate the Flux prompt for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeFluxPrompt(string $search): string
    {
        $search = Str::lower($search);
        $customInstructions = $this->getCustomInstructions();

        $preset = Preset::findFromAPI("generate_flux_prompt");

        $response = $preset->process([
            "request" => $search,
            "has_custom_instructions" => !empty($customInstructions),
            "custom_instructions" => $customInstructions,
        ]);

        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        if (!empty($content)) {
            return $content;
        }

        return $search;
    }

    /**
     * Generate the ideogram prompt for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeIdeogramPrompt(string $search): string
    {
        $search = Str::lower($search);
        $customInstructions = $this->getCustomInstructions();

        $preset = Preset::findFromAPI("generate_ideogram_prompt");

        $response = $preset->process([
            "request" => $search,
            "has_custom_instructions" => !empty($customInstructions),
            "custom_instructions" => $customInstructions,
        ]);

        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        if (!empty($content)) {
            return $content;
        }

        return $search;
    }

    /**
     * Generate the dall-e prompt for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeDallEPrompt(string $search): string
    {
        $search = Str::lower($search);
        $customInstructions = $this->getCustomInstructions();

        $preset = Preset::findFromAPI("generate_dall_e_prompt");

        $response = $preset->process([
            "language" => $this->publication->project->language->value,
            "request" => $search,
            "has_custom_instructions" => !empty($customInstructions),
            "custom_instructions" => $customInstructions,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        // custom instructions
        $customInstructions =  $this->getCustomInstructions();
        if (!empty($customInstructions)) {
            $content = $content . " " . $customInstructions;
        }

        return Str::lower($content);
    }

    /**
     * Make the Unsplash search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeUnsplashSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("get_main_keyword");

        $response = $preset->process([
            "language" => Language::find("en")->value,
            "request" => $search,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Make the Pexels search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makePexelsSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("get_main_keyword");

        $response = $preset->process([
            "language" => Language::find("en")->value,
            "request" => $search,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Make the Pixabay search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makePixabaySearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("get_main_keyword");

        $response = $preset->process([
            "language" => Language::find("en")->value,
            "request" => $search,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Make the Google Image search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeGoogleImageSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("shorten_title");

        $response = $preset->process([
            "language" => $this->publication->project->language->value,
            "request" => $search,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Make the Bong Image search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeBingImageSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("shorten_title");

        $response = $preset->process([
            "language" => $this->publication->project->language->value,
            "request" => $search,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Generate an image description from the title
     *
     * @param string $title
     * @return string
     * @throws Exception
     */
    private function makeImageDescription(string $title): string
    {
        $preset = Preset::findFromAPI("generate_alt_image");

        $response = $preset->process([
            "language" => $this->publication->project->language->value,
            "request" => $title,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Get custom instructions
     *
     * @return string
     */
    private function getCustomInstructions()
    {
        return Arr::get($this->publication->project->modules, "image.custom_instructions", "");
    }

    /**
     * Get the number of Image blocks
     *
     * @param Publication $publication
     * @return integer
     */
    static private function getLength(Publication $publication): int
    {
        $length = 0;

        if (Arr::get($publication->project->modules, "image.thumbnail.enabled", false)) {
            $length++;
        }

        foreach (Arr::get($publication->project->modules, 'blocks', []) as $block) {
            if ($block["block"] === "image") {
                $length++;
            }
        }

        return $length;
    }

    /**
     * Determine if the module is enabled
     *
     * @param Publication $publication
     * @return boolean
     */
    static public function isEnabled(Publication $publication): bool
    {
        return self::getLength($publication) > 0;
    }

    /**
     * Determine if the module is runnable
     *
     * @param Publication $publication
     * @return boolean
     */
    static public function isRunnable(Publication $publication): bool
    {
        // must be enabled
        if (!self::isEnabled($publication)) {
            return false;
        }

        return true;
    }
}
