<?php

namespace OtomaticAi\Modules;

use Exception;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Content\Social\Instagram;
use OtomaticAi\Models\Presets\Preset;
use OtomaticAi\Modules\Contracts\Module as ModuleContract;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;

class ProcessInstagramModule extends Module implements ModuleContract
{
    /**
     * Execute the job.
     *
     * @return void
     * @throws Exception
     */
    public function handle(): void
    {
        // verify that the job is runnable
        if (!$this->isRunnable()) {
            return;
        }

        // log the start of the job
        $this->publication->addLog("Instagram module started.");

        // get length
        $length = $this->getLength();

        // create instagram placeholders
        $placeholders = $this->createElementPlaceholders("instagram");

        try {
            // make the search terms
            $search = $this->makeSearch($this->getTitle());

            // call the api
            $api = new Client;
            $result = $api->instagram($search, $this->getLanguage()->key);

            // get the organic results
            $posts = is_array($result) ? $result : [];

            // randomize posts
            $posts = Arr::shuffle($posts);

            if (!empty($posts)) {

                // get x random posts
                $posts = Arr::random($posts, min(count($posts), $length));
                $posts = array_map(function ($post) {
                    return new Instagram($post["link"]);
                }, $posts);

                // replace instagram placeholders with the generated posts
                $this->replaceElementPlaceholders("instagram", function ($attributes) use ($posts) {
                    $index = Arr::get($attributes, "index");
                    if ($index !== null && isset($posts[$index])) {
                        $post = $posts[$index];

                        $el = $post->toHtmlElement($this->html);
                        if ($el) {
                            return $el;
                        }
                    }
                });

                // log the end of the job
                $this->publication->addLog("Instagram module completed successfully.", "success");
            } else {
                $this->publication->addLog("Instagram module failed. No posts found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Instagram module failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Make the Instagram search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("get_main_keyword");

        $response = $preset->process([
            "language" => $this->getLanguage()->value,
            "title" => $this->getTitle(),
            "request" => $this->getTitle(),
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Get the number of Instagram elements
     *
     * @return integer
     */
    private function getLength(): int
    {
        return $this->countElements("instagram");;
    }

    /**
     * Determine if the module is runnable
     *
     * @return boolean
     */
    public function isRunnable(): bool
    {
        // must have instagram elements
        return $this->getLength() > 0;
    }
}
