<?php

namespace OtomaticAi\Modules;

use Exception;
use OtomaticAi\Content\Section;
use OtomaticAi\Modules\Contracts\Module as ModuleContract;
use OtomaticAi\Models\Publication;
use OtomaticAi\Models\Preset;
use OtomaticAi\Utils\GoogleNews;
use OtomaticAi\Utils\HtmlParser;
use OtomaticAi\Utils\RobotsTxt;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;
use OtomaticAi\Vendors\andreskrey\Readability\Readability;
use OtomaticAi\Vendors\andreskrey\Readability\Configuration;
use OtomaticAi\Api\OtomaticAi\Client as OtomaticAiClient;
use OtomaticAi\Utils\PresetCollection;
use OtomaticAi\Vendors\Carbon\Carbon;
use WP_Query;

class ProcessTextModule implements ModuleContract
{
    /**
     * The publication
     *
     * @var Publication
     */
    private Publication $publication;

    private $persona;

    /**
     * Create a new job instance.
     *
     * @param Publication $publication
     */
    public function __construct(Publication $publication)
    {
        $this->publication = $publication;
    }

    /**
     * Execute the job.
     *
     * @return void
     * @throws Exception
     */
    public function handle(): void
    {
        // verify that the job is runnable
        if (!self::isRunnable($this->publication)) {
            return;
        }

        // log the start of the job
        $this->publication->addLog("Text module started.");

        // perform the correct text model
        if ($this->publication->project->type === "news" || $this->publication->project->type === "news-now") {
            $this->createNewsSections();
        } else if ($this->publication->project->type === "rss" || $this->publication->project->type === "rss-now") {
            $this->createRssSections();
        } else if ($this->publication->project->type === "sitemap") {
            $this->createSitemapSections();
        } else if ($this->publication->project->type === "url") {
            $this->createUrlSections();
        } else if ($this->publication->project->type === "youtube") {
            $this->createYoutubeSections();
        } else {
            $this->createSingleModeSections();
        }

        // log the end of the job
        if ($this->publication->sections->isNotEmpty()) {
            $this->publication->addLog("Text module completed successfully.", "success");
        } else {
            $this->publication->addLog("Text module completed with errors. No content generated.", "error");
        }
    }

    /**
     * Generate news sections
     *
     * @return void
     * @throws Exception
     */
    private function createNewsSections()
    {
        // get the meta.url
        $url = Arr::get($this->publication->meta, 'url');
        if (empty($url)) {
            throw new Exception("News url is empty.");
        }

        $this->createScrapingSections($url);
    }

    /**
     * Generate rss sections
     *
     * @return void
     * @throws Exception
     */
    private function createRssSections()
    {
        // get the meta.url
        $url = Arr::get($this->publication->meta, 'url');
        if (empty($url)) {
            throw new Exception("Rss url is empty.");
        }

        $this->createScrapingSections($url);
    }

    /**
     * Generate sitemap sections
     *
     * @return void
     * @throws Exception
     */
    private function createSitemapSections()
    {
        // get the meta.url
        $url = Arr::get($this->publication->meta, 'url');
        if (empty($url)) {
            throw new Exception("Url is empty.");
        }

        $this->createScrapingSections($url);
    }

    /**
     * Generate url sections
     *
     * @return void
     * @throws Exception
     */
    private function createUrlSections()
    {
        // get the meta.url
        $url = Arr::get($this->publication->meta, 'url');
        if (empty($url)) {
            throw new Exception("Url is empty.");
        }

        $this->createScrapingSections($url);
    }

    /**
     * Generate youtube sections
     *
     * @return void
     * @throws Exception
     */
    private function createYoutubeSections()
    {
        // get the meta.url
        $videoId = Arr::get($this->publication->meta, 'videoId');
        if (empty($videoId)) {
            throw new Exception("YouTube video ID is empty.");
        }

        $captions = null;
        try {
            $api = new OtomaticAiClient;
            $result = $api->youtubeVideo($videoId, $this->publication->project->language->key);
            $captions = Arr::get($result, "captions");
        } catch (Exception $e) {
            throw new Exception("Failed to get the YouTube video captions.");
        }

        if (empty($captions)) {
            throw new Exception("YouTube video captions are empty.");
        }

        $this->createSingleModeSections($captions);
    }

    private function createScrapingSections(string $url)
    {
        // verify robots txt
        if ($this->isDisabledByRobotsTxt($url)) {
            throw new Exception("Disabled by robots.txt.");
        }

        // scrap the meta.url content
        $response = wp_remote_get($url, [
            "headers" => [
                "User-Agent" => "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
            ]
        ]);

        // quit if the content is not usable
        if (is_wp_error($response) || empty($body = Arr::get($response, "body")) || strlen($body) < 200) {
            throw new Exception("Failed to get the url content.");
        }

        $content = null;
        try {
            // remove boilerplate frop body
            $configuration = new Configuration([
                "articleByLine" => true,
                "summonCthulhu" => true,
                "normalizeEntities" => true,
            ]);
            $readability = new Readability($configuration);
            $readability->parse($body);
            $content = $readability->getContent();
        } catch (Exception $e) {
        }

        // quit if the content is empty
        if (empty($content)) {
            return;
        }

        $this->createSingleModeSections($content);
    }

    /**
     * Generate single mode sections
     *
     * @param string|null $scrapedContent
     * @return void
     * @throws Exception
     */
    private function createSingleModeSections(string $scrapedContent = null)
    {
        // get publication sections
        $sections = $this->publication->sections;

        // make payload
        $payload = $this->makeInitialPayload();

        // get external and internal data
        $externalData = $this->getExternalData();
        $internalData = $this->getInternalData();

        // get external links and randomize
        $externalLinks = Arr::get($externalData, "links", []);
        $perBlockExternalLinks = min(ceil(count($externalLinks) / max(1, count($this->getExternalLinksEnabledBlocks()))), 5);
        shuffle($externalLinks);

        // get internal links and randomize
        $internalLinks = Arr::get($internalData, "links", []);
        $perBlockInternalLinks = min(ceil(count($internalLinks) / max(1, count($this->getInternalLinksEnabledBlocks()))), 5);
        shuffle($internalLinks);

        $blocksPresets = [];

        // get the system
        $textBlockSystem = $this->getTextBlockSystem();

        // for each blocks set by the user
        foreach ($this->getBlocks() as $blockIndex => $block) {

            if ($block["block"] === "text") {
                $prompt = Arr::get($block, "settings.prompt", "");
                if (empty($prompt)) {
                    continue;
                }

                // create the block preset
                $preset = Preset::make([
                    "model" => self::getAIModel($this->publication),
                    "messages" => [
                        [
                            "role" => "system",
                            "content" => $textBlockSystem,
                        ],
                        [
                            "role" => "user",
                            "content" => $prompt,
                        ]
                    ],
                ]);

                // create the data variable with scraped content, scraped external data and internal data
                $data = [];
                if (!empty($scrapedContent)) {
                    $data[] = $scrapedContent;
                }
                if (Arr::get($block, "settings.use_live_data", false)) {
                    $data = array_merge($data, Arr::get($externalData, "contents", []), Arr::get($internalData, "contents", []));
                }
                $data = implode("\n", $data);

                // create the links variable with scraped external data and internal data
                $links = [];
                if (Arr::get($block, "settings.external_links_enabled", false)) {
                    $links = array_merge($links, array_splice($externalLinks, 0, $perBlockExternalLinks));
                }
                if (Arr::get($block, "settings.internal_links_enabled", false)) {
                    $links = array_merge($links, array_splice($internalLinks, 0, $perBlockInternalLinks));
                }
                shuffle($links);
                $links = implode("\n", $links);

                // add to the array of presets
                $blocksPresets[] = [
                    $preset,
                    array_merge($payload, [
                        "has_data" => !empty($data),
                        "data" => $data,
                        "has_links" => !empty($links),
                        "links" => $links,
                    ]),
                    function (array $response) use (&$sections, $blockIndex) {
                        // callback of the preset 

                        // get the response content
                        $content = Arr::get($response, 'choices.0.message.content');

                        if ($content) {
                            // parse content
                            $html = new HtmlParser($content);

                            // create a new section with the html and add it to the section collection
                            if (!empty($html)) {
                                $section = new Section();
                                $section->addChild($html->getSections([
                                    "blacklist_titles" => [$this->publication->generation_title]
                                ]));

                                $sections->put($blockIndex, $section);
                            }
                        }
                    }
                ];
            }
        }

        // run the block preset collection
        if (!empty($blocksPresets)) {
            PresetCollection::run($blocksPresets);
        }

        $this->publication->sections = $sections;
    }

    /**
     * Create the default payload
     *
     * @return array
     */
    private function makeInitialPayload(): array
    {
        // persona
        $persona = $this->getPersona();

        // buyer persona
        $buyerPersona = $this->getBuyerPersona();

        // writing style
        $writingStyle = $this->getWritingStyle();

        // make payload
        $payload = [
            "language" => $this->publication->project->language->value,
            "title" => $this->publication->generation_title,
            "request" => $this->publication->generation_title,
            "has_persona" => !empty($persona),
            "persona" => $persona,
            "has_buyer_persona" => !empty($buyerPersona),
            "buyer_persona" => $buyerPersona,
            "has_writing_style" => !empty($writingStyle),
            "writing_style" => $writingStyle,
        ];

        return $payload;
    }

    /**
     * Undocumented function
     *
     * @return array
     */
    private function getExternalData(): array
    {
        $data = [
            "contents" => [],
            "links" => [],
        ];

        if (count($this->getUseDataBlocks()) > 0 || count($this->getExternalLinksEnabledBlocks()) > 0) {

            // get google results
            $googleResults = $this->getGoogleResults($this->publication->generation_title);

            // organic results
            $organicResults = Arr::get($googleResults, "organic_results", []);
            if (count($organicResults) > 0) {
                $organicResults = Arr::random($organicResults, min(count($organicResults), 10));
                $organicResults = array_map(function ($item) {
                    return [
                        "link" => Arr::get($item, "link", ""),
                        "content" => trim(Arr::get($item, "title", "") . "\n" . Arr::get($item, "snippet", ""))
                    ];
                }, $organicResults);

                $data["contents"] = array_merge($data["contents"], array_map(function ($item) {
                    return $item["content"];
                }, $organicResults));
                $data["links"] = array_merge($data["links"], array_map(function ($item) {
                    return $item["link"];
                }, $organicResults));
            }

            // related questions
            $relatedQuestions = Arr::get($googleResults, "related_questions", []);
            if (count($relatedQuestions) > 0) {
                $relatedQuestions = Arr::random($relatedQuestions, min(count($relatedQuestions), 10));
                $relatedQuestions = array_map(function ($item) {
                    $rawSnippet = Arr::get($item, "snippet", []);
                    $snippet = [];
                    if (Arr::has($rawSnippet, "heading")) {
                        $snippet[] = $rawSnippet["heading"];
                    }
                    if (Arr::has($snippet, "list")) {
                        $snippet = array_merge($snippet, $snippet["list"]);
                    }
                    if (Arr::has($snippet, "table")) {
                        foreach ($snippet["table"] as $row) {
                            $snippet = array_merge($snippet, $row);
                        }
                    }
                    if (Arr::has($snippet, "text")) {
                        $snippet[] = $snippet["text"];
                    }

                    $snippet = implode("\n", $snippet);

                    return [
                        "link" => Arr::get($item, "link", ""),
                        "content" => trim(Arr::get($item, "question", "") . "\n" . Arr::get($item, "title", "") . "\n" .  $snippet)
                    ];
                }, $relatedQuestions);

                $data["contents"] = array_merge($data["contents"], array_map(function ($item) {
                    return $item["content"];
                }, $relatedQuestions));
                $data["links"] = array_merge($data["links"], array_map(function ($item) {
                    return $item["link"];
                }, $relatedQuestions));
            }

            // related searches
            $relatedSearches = Arr::get($googleResults, "related_searches", []);
            if (count($relatedSearches) > 0) {
                $relatedSearches = Arr::random($relatedSearches, min(count($relatedSearches), 10));
                $relatedSearches = array_map(function ($item) {
                    return [
                        "content" => trim(Arr::get($item, "query", "")),
                    ];
                }, $relatedSearches);

                $data["contents"] = array_merge($data["contents"], array_map(function ($item) {
                    return $item["content"];
                }, $relatedSearches));
            }

            // get google news links
            $newsLinks = $this->getGoogleNewsLinks($this->publication->generation_title);
            if (count($newsLinks) > 0) {
                $data["contents"] = array_merge($data["contents"], array_map(function ($item) {
                    return trim($item["title"]);
                }, $newsLinks));
                $data["links"] = array_merge($data["links"], array_map(function ($item) {
                    return $item["url"];
                }, $newsLinks));
            }
        }

        return $data;
    }

    /**
     * Undocumented function
     *
     * @return array
     */
    private function getInternalData(): array
    {
        $data = [
            "contents" => [],
            "links" => [],
        ];

        if (count($this->getInternalLinksEnabledBlocks()) > 0) {

            $relatedPosts = $this->getRelatedPosts();
            if (!empty($relatedPosts)) {
                $relatedPosts = array_map(function ($post) {
                    return [
                        "title" => get_the_title($post),
                        "link" => get_permalink($post),
                    ];
                }, $relatedPosts);

                $data["contents"] = array_merge($data["contents"], array_map(function ($item) {
                    return $item["title"];
                }, $relatedPosts));
                $data["links"] = array_merge($data["links"], array_map(function ($item) {
                    return $item["link"];
                }, $relatedPosts));
            }
        }

        return $data;
    }

    /**
     * Undocumented function
     *
     * @return array
     */
    private function getUseDataBlocks(): array
    {
        return array_values(array_filter($this->getBlocks(), function ($block) {
            return $block["block"] === "text" && Arr::get($block, "settings.use_live_data", false);
        }));
    }

    /**
     * Undocumented function
     *
     * @return array
     */
    private function getExternalLinksEnabledBlocks(): array
    {
        return array_values(array_filter($this->getBlocks(), function ($block) {
            return $block["block"] === "text" && Arr::get($block, "settings.internal_links_enabled", false);
        }));
    }

    /**
     * Undocumented function
     *
     * @return array
     */
    private function getInternalLinksEnabledBlocks(): array
    {
        return array_values(array_filter($this->getBlocks(), function ($block) {
            return $block["block"] === "text" && Arr::get($block, "settings.internal_links_enabled", false);
        }));
    }

    /**
     * Undocumented function
     *
     * @return array
     */
    private function getBlocks(): array
    {
        return Arr::get($this->publication->project->modules, 'blocks', []);
    }

    /**
     * Fetch Google News items
     *
     * @param string $search
     * @return array
     */
    private function getGoogleNewsLinks(string $search): array
    {
        if (!Arr::get($this->publication->project->modules, 'text.external_links.enabled', false)) {
            return [];
        }

        // retrieve google news rss
        try {
            return GoogleNews::search($search, $this->publication->project->language);
        } catch (Exception $e) {
            return [];
        }
    }

    private function getGoogleResults(string $search)
    {
        // retrieve google news rss
        try {
            $api = new OtomaticAiClient;
            return $api->googleSearch($search, $this->publication->project->language->key);
        } catch (Exception $e) {
            return [];
        }
    }

    private function getRelatedPosts()
    {
        $search = Str::lower($this->publication->generation_title);

        try {
            $preset = Preset::findFromAPI("get_main_keyword");

            $response = $preset->process([
                "language" => $this->publication->project->language->value,
                "request" => $search,
            ]);

            // get the response content
            $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
            $content = Arr::get($response, "value");
            $content = Str::clean($content);

            $search = Str::lower($content);
        } catch (Exception $e) {
        }

        if (empty($search)) {
            return [];
        }

        $args = [
            "s" => $search,
        ];

        if (!empty($this->publication->post_id)) {
            $args["post__not_in"] = [$this->publication->post_id];
        }
        $query = new WP_Query($args);

        return $query->get_posts();
    }

    /**
     * Get the persona system message form selected, or generate one
     *
     * @return string
     * @throws Exception
     */
    private function getPersona()
    {
        if ($this->publication->project->persona) {
            return $this->publication->project->persona->openai_system_message;
        } else if (!empty($this->persona)) {
            return $this->persona;
        }

        $this->persona = $this->createPersona();

        return $this->persona;
    }

    /**
     * Get the buyer persona
     *
     * @return string
     */
    private function getBuyerPersona()
    {
        return Arr::get($this->publication->project->modules, "text.buyer_persona.description", "");
    }

    /**
     * Get the writing style
     *
     * @return string
     */
    private function getWritingStyle()
    {
        switch (Arr::get($this->publication->project->modules, "text.writing_style")) {
            case "narrative":
                return "Narratif";
                break;
            case "descriptive":
                return "Descriptif";
                break;
            case "objective":
                return "Objectif";
                break;
            case "poetic":
                return "Poétique";
                break;
            case "persuasive":
                return "Persuasif";
                break;
            case "expository":
                return "Expositoire";
                break;
            case "creative":
                return "Créatif";
                break;
            case "technical":
                return "Technique";
                break;
            case "entertaining":
                return "Divertissant";
                break;
            case "subjective":
                return "Subjectif";
                break;
            case "funny":
                return "Drôle";
                break;
            case "sales-oriented":
                return "Vendeur";
                break;
            case "formal":
                return "Soutenu";
                break;
            case "ironic":
                return "Ironique";
                break;
            case "inspiring":
                return "Inspirant";
                break;
            case "educative":
                return "Éducatif";
                break;
            case "motivating":
                return "Motivant";
                break;
            case "informal":
                return "Informel";
                break;
            case "provocative":
                return "Provocateur";
                break;
            case "emotional":
                return "Émotionnel";
                break;
            case "critical":
                return "Critique";
                break;
            case "sarcastic":
                return "Sarcastique";
                break;
            case "imaginative":
                return "Imaginatif";
                break;
            case "pragmatic":
                return "Pragmatique";
                break;
            case "argumentative":
                return "Argumentatif";
                break;
            case "factual":
                return "Factuel";
                break;
        }

        return "";
    }

    /**
     * Undocumented function
     *
     * @return string
     */
    private function getTextBlockSystem(): string
    {
        $default = implode("\n", [
            "=====================",
            "Instructions de génération:",
            "=====================",
            "La sortie du texte doit être en \"[LANGUAGE]\" et en HTML. C'est très important.",
            "---",
            "[IF;HAS_PERSONA]",
            "Ce texte est rédigé par: [PERSONA]",
            "Ne cite pas le persona qui a ecrit ce texte dans le paragraphe.",
            "---",
            "[ENDIF]",
            "[IF;HAS_BUYER_PERSONA]",
            "Ce texte cible ce type de buyer persona, voici un exemple de buyer persona: [BUYER_PERSONA]",
            "Important : ce n'est qu'un exemple ne buyer persona. Ne reprend et ne cite pas  le persona ou le nom cité dans l'exemple.",
            "---",
            "[ENDIF]",
            "[IF;HAS_WRITING_STYLE]",
            "Ecrit avec ce style d'écriture : [WRITING_STYLE]",
            "---",
            "[ENDIF]",
            "[IF;HAS_DATA]",
            "Utilise des élément de ce texte pour alimenter l'article. Il est important que l'article soit différent du texte source, en changeant au maximum e style du texte: ",
            "[DATA]",
            "---",
            "[ENDIF]",
            "[IF;HAS_LINKS]",
            "Inclu ces lien dans le texte de, maniere naturel.",
            "[LINKS]",
            "---",
            "[ELSE]",
            "N'ajoute pas de lien externe (https// ...)  dans le texte ! ",
            "---",
            "[ENDIF]",
            "Enlève le formatage Markdown du texte, ne décrit pas ce que tu génères.",
            "---",
            "Highlight important keywords in <strong>.",
            "=====================",
        ]);

        $optionName = "otomatic_ai_text_block_system";

        $output = [];

        if (function_exists('get_option')) {
            $cache = get_option($optionName, null);

            // fetch or refresh cache
            if (
                empty($cache) ||
                Carbon::createFromFormat(
                    "Y-m-d H:i:s",
                    Arr::get(
                        $cache,
                        "cached_at",
                        Carbon::now()->subMonth()
                    )
                )
                ->lt(
                    Carbon::now()->subMinutes(5)
                )
            ) {
                try {
                    $api = new OtomaticAiClient;
                    $output = $api->textBlockSystem(OTOMATIC_AI_VERSION);
                    $output = Arr::get($output, "prompt", $default);

                    $cache = [
                        "data" => $output,
                        "cached_at" => Carbon::now()->format("Y-m-d H:i:s"),
                    ];

                    if (function_exists('update_option')) {
                        update_option($optionName, $cache, false);
                    }
                } catch (Exception $e) {
                    $output = $default;

                    $cache = [
                        "data" => $output,
                        "cached_at" => Carbon::now()->format("Y-m-d H:i:s"),
                    ];

                    if (function_exists('update_option')) {
                        update_option($optionName, $cache, false);
                    }
                }
            } else {
                $output = Arr::get($cache, "data", $output);
            }
        }

        return $output;
    }

    /**
     * Create an automatic persona
     *
     * @return string
     * @throws Exception
     */
    private function createPersona()
    {
        // get the persona preset
        $preset = Preset::findFromAPI("generate_persona");

        // run the preset
        $response = $preset->process([
            "language" => $this->publication->project->language->value,
            "request" => $this->publication->generation_title,
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");

        return Str::clean($content);
    }

    /**
     * Determine if the scrap is disabled by the robots.txt
     *
     * @param string $url
     * @return boolean
     */
    private function isDisabledByRobotsTxt(string $url)
    {
        $robots = RobotsTxt::fromUrl($url);
        return !empty($robots->getRules("otomaticAI"));
    }

    /**
     * return the ai model key to use
     *
     * @param Publication $publication
     * @param string $key
     * @return string
     */
    static public function getAIModel(Publication $publication): string
    {
        return Arr::get($publication->project->modules, "text.model", 'gpt-4o-mini');
    }

    /**
     * Get the number of Text blocks
     *
     * @param Publication $publication
     * @return integer
     */
    static private function getLength(Publication $publication): int
    {
        $length = 0;

        foreach (Arr::get($publication->project->modules, 'blocks', []) as $block) {
            if ($block["block"] === "text") {
                $length++;
            }
        }

        return $length;
    }

    /**
     * Determine if the module is enabled
     *
     * @param Publication $publication
     * @return boolean
     */
    static public function isEnabled(Publication $publication): bool
    {
        return self::getLength($publication) > 0;
    }

    /**
     * Determine if the module is runnable
     *
     * @param Publication $publication
     * @return boolean
     */
    static public function isRunnable(Publication $publication): bool
    {
        // must be enabled
        if (!self::isEnabled($publication)) {
            return false;
        }

        return true;
    }
}
