<?php

namespace OtomaticAi\Modules;

use Exception;
use OtomaticAi\Vendors\GuzzleHttp\Client as GuzzleHttpClient;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Content\Section;
use OtomaticAi\Content\Social\Twitter;
use OtomaticAi\Models\Presets\Preset;
use OtomaticAi\Modules\Contracts\Module as ModuleContract;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;

class ProcessTwitterModule extends Module implements ModuleContract
{
    /**
     * Execute the job.
     *
     * @return void
     * @throws Exception
     */
    public function handle(): void
    {
        // verify that the job is runnable
        if (!$this->isRunnable()) {
            return;
        }

        // log the start of the job
        $this->publication->addLog("Twitter module started.");

        // get length
        $length = $this->getLength();

        // create twitter placeholders
        $placeholders = $this->createElementPlaceholders("twitter");

        try {
            // make the search terms
            $search = $this->makeSearch($this->getTitle());

            // call the api
            $api = new Client;
            $result = $api->twitter($search, $this->getLanguage()->key);

            // get the organic results
            $posts = is_array($result) ? $result : [];

            // randomize posts
            $posts = Arr::shuffle($posts);

            // transform mobile link into desktop link
            $posts = array_map(function ($post) {
                $post["link"] = preg_replace('/^https?:\/\/(mobile.)/', 'https://', $post["link"]);
                return $post;
            }, $posts);

            // verify if the post is displayable
            $posts = array_values(array_filter($posts, function ($post) {
                return $this->isPostDisplayable($post["link"]);
            }));

            if (!empty($posts)) {

                // get x random posts
                $posts = Arr::random($posts, min(count($posts), $length));
                $posts = array_map(function ($post) {
                    return new Twitter($post["link"]);
                }, $posts);

                // replace twitter placeholders with the generated posts
                $this->replaceElementPlaceholders("twitter", function ($attributes) use ($posts) {
                    $index = Arr::get($attributes, "index");
                    if ($index !== null && isset($posts[$index])) {
                        $post = $posts[$index];

                        $el = $post->toHtmlElement($this->html);
                        if ($el) {
                            return $el;
                        }
                    }
                });

                // log the end of the job
                $this->publication->addLog("Twitter module completed successfully.", "success");
            } else {
                $this->publication->addLog("Twitter module failed. No posts found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Twitter module failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Make the Twitter search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeSearch(string $search): string
    {
        $search = Str::lower($search);

        $preset = Preset::findFromAPI("get_main_keyword");

        $response = $preset->process([
            "language" => $this->getLanguage()->value,
            "request" => $this->getTitle(),
        ]);

        // get the response content
        $response = json_decode(Arr::get($response, 'choices.0.message.content'), true, 512, JSON_THROW_ON_ERROR);
        $content = Arr::get($response, "value");
        $content = Str::clean($content);

        return Str::lower($content);
    }

    /**
     * Determine if the link is displayble
     *
     * @param string $link
     * @return boolean
     */
    private function isPostDisplayable(string $link): bool
    {
        try {
            $client = new GuzzleHttpClient();
            $client->get("https://publish.twitter.com/oembed", [
                "query" => [
                    "url" => $link,
                ]
            ]);
            return true;
        } catch (Exception $e) {
            return false;
        }
    }

    /**
     * Get the number of Twitter elements
     *
     * @return integer
     */
    private function getLength(): int
    {
        return $this->countElements("twitter");;
    }

    /**
     * Determine if the module is runnable
     *
     * @return boolean
     */
    public function isRunnable(): bool
    {
        // must have twitter elements
        return $this->getLength() > 0;
    }
}
