<?php

namespace OtomaticAi\Modules;

use Exception;
use OtomaticAi\Api\OtomaticAi\Client;
use OtomaticAi\Content\Section;
use OtomaticAi\Content\Video\Youtube;
use OtomaticAi\Models\Publication;
use OtomaticAi\Modules\Contracts\Module as ModuleContract;
use OtomaticAi\Vendors\Illuminate\Support\Arr;
use OtomaticAi\Vendors\Illuminate\Support\Str;

class ProcessYoutubeModule implements ModuleContract
{
    /**
     * The publication
     *
     * @var Publication
     */
    private Publication $publication;

    /**
     * Create a new job instance.
     *
     * @param Publication $publication
     */
    public function __construct(Publication $publication)
    {
        $this->publication = $publication;
    }

    /**
     * Execute the job.
     *
     * @return void
     * @throws Exception
     */
    public function handle(): void
    {
        // verify that the job is runnable
        if (!self::isRunnable($this->publication)) {
            return;
        }

        // log the start of the job
        $this->publication->addLog("Youtube module started.");

        // determine length
        $length = self::getLength($this->publication);

        try {
            // make the search terms
            $search = $this->makeSearch($this->publication->generation_title);

            // call the api
            $api = new Client;
            $result = $api->youtube($search, $this->publication->project->language->key);

            // get the videos
            $videos = is_array($result) ? $result : [];

            // randomize videos
            $videos = Arr::shuffle($videos);

            if (!empty($videos)) {

                // get x random posts
                $videos = Arr::random($videos, min(count($videos), $length));
                $videos = array_map(function ($video) {
                    return  new Youtube($video["id"], $video["title"]);
                }, $videos);

                // get publication sections
                $sections = $this->publication->sections;

                // add posts to sections
                $index = 0;
                foreach (Arr::get($this->publication->project->modules, 'blocks', []) as $blockIndex => $block) {
                    if ($block["block"] === "youtube" && isset($videos[$index])) {
                        $section = new Section();
                        $section->addElement($videos[$index]);
                        $sections->put($blockIndex, $section);
                        $index++;
                    }
                }

                // update publication sections
                $this->publication->sections = $sections;

                // log the end of the job
                $this->publication->addLog("Youtube module completed successfully.", "success");
            } else {
                $this->publication->addLog("Youtube module failed. No videos found.", "warning");
            }
        } catch (Exception $e) {
            $this->publication->addLog("Youtube module failed. " . $e->getMessage(), "warning");
        }
    }

    /**
     * Make the Youtube search terms for the provided search
     *
     * @param string $search
     * @return string
     * @throws Exception
     */
    private function makeSearch(string $search): string
    {
        $search = Str::lower($search);

        return $search;
    }

    /**
     * Get the number of Youtube blocks
     *
     * @param Publication $publication
     * @return integer
     */
    static private function getLength(Publication $publication): int
    {
        $length = 0;

        foreach (Arr::get($publication->project->modules, 'blocks', []) as $block) {
            if ($block["block"] === "youtube") {
                $length++;
            }
        }

        return $length;
    }

    /**
     * Determine if the module is enabled
     *
     * @param Publication $publication
     * @return boolean
     */
    static public function isEnabled(Publication $publication): bool
    {
        return self::getLength($publication) > 0;
    }

    /**
     * Determine if the module is runnable
     *
     * @param Publication $publication
     * @return boolean
     */
    static public function isRunnable(Publication $publication): bool
    {
        // must be enabled
        if (!self::isEnabled($publication)) {
            return false;
        }

        return true;
    }
}
